"""
This module defines the wrapper for ML engines which abstracts away a lot of complexity.

In particular, three big components are included:

    - `BaseMLEngineExec` class: this class wraps any object that inherits from `BaseMLEngine` and exposes some endpoints
      normally associated with a DB handler (e.g. `native_query`, `get_tables`), as well as other ML-specific behaviors,
      like `learn()` or `predict()`. Note that while these still have to be implemented at the engine level, the burden
      on that class is lesser given that it only needs to return a pandas DataFrame. It's this class that will take said
      output and format it into the HandlerResponse instance that MindsDB core expects.

    - `learn_process` method: handles async dispatch of the `learn` method in an engine, as well as registering all
      models inside of the internal MindsDB registry.

    - `predict_process` method: handles async dispatch of the `predict` method in an engine.

"""

import socket
import datetime as dt
from typing import Optional

import pandas as pd
from sqlalchemy import func, null
from sqlalchemy.sql.functions import coalesce

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities.config import Config
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.utilities.hooks import after_predict as after_predict_hook
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.model.functions import (
    get_model_record
)
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.model.functions import get_model_records
from mindsdb.utilities.functions import mark_process
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.ml_task_queue.producer import MLTaskProducer
from mindsdb.utilities.ml_task_queue.const import ML_TASK_TYPE
from mindsdb.integrations.libs.process_cache import process_cache, empty_callback

try:
    import torch.multiprocessing as mp
except Exception:
    import multiprocessing as mp
mp_ctx = mp.get_context('spawn')


class MLEngineException(Exception):
    pass


class BaseMLEngineExec:

    def __init__(self, name, integration_id, integration_engine, handler_class):
        """ ML handler interface converter

            Args:
                name
                integration_id
                integration_engine
                handler_class
        """  # noqa
        # TODO move this class to model controller

        self.name = name
        self.config = Config()
        self.integration_id = integration_id
        self.engine = integration_engine
        self.handler_class = handler_class

        self.model_controller = ModelController()
        self.database_controller = DatabaseController()

        self.parser = parse_sql
        self.dialect = 'mindsdb'

        self.is_connected = True

        self.base_ml_executor = process_cache
        if self.config['ml_task_queue']['type'] == 'redis':
            self.base_ml_executor = MLTaskProducer()

    def get_ml_handler(self, predictor_id=None):
        # returns instance or wrapper over it

        integration_id = self.integration_id

        handlerStorage = HandlerStorage(integration_id)
        modelStorage = ModelStorage(predictor_id)

        ml_handler = self.handler_class(
            engine_storage=handlerStorage,
            model_storage=modelStorage,
        )
        return ml_handler

    def get_tables(self) -> Response:
        """ Returns all models currently registered that belong to the ML engine."""
        all_models = self.model_controller.get_models(integration_id=self.integration_id)
        all_models_names = [[x['name']] for x in all_models]
        response = Response(
            RESPONSE_TYPE.TABLE,
            pd.DataFrame(
                all_models_names,
                columns=['table_name']
            )
        )
        return response

    def get_columns(self, table_name: str) -> Response:
        """ Retrieves standard info about a model, e.g. data types. """  # noqa
        predictor_record = get_model_record(name=table_name, ml_handler_name=self.name)
        if predictor_record is None:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message=f"Error: model '{table_name}' does not exist!"
            )

        data = []
        if predictor_record.dtype_dict is not None:
            for key, value in predictor_record.dtype_dict.items():
                data.append((key, value))
        result = Response(
            RESPONSE_TYPE.TABLE,
            pd.DataFrame(
                data,
                columns=['COLUMN_NAME', 'DATA_TYPE']
            )
        )
        return result

    def native_query(self, query: str) -> Response:
        """ Intakes a raw SQL query and returns the answer given by the ML engine. """
        query_ast = self.parser(query, dialect=self.dialect)
        return self.query(query_ast)

    def query_(self, query: ASTNode) -> Response:
        raise Exception('Should not be used')

    @profiler.profile()
    def learn(
        self, model_name, project_name,
        data_integration_ref=None,
        fetch_data_query=None,
        problem_definition=None,
        join_learn_process=False,
        label=None,
        is_retrain=False,
        set_active=True,
    ):
        # TODO move to model_controller
        """ Trains a model given some data-gathering SQL statement. """

        # may or may not be provided (e.g. 0-shot models do not need it), so engine will handle it
        target = problem_definition.get('target', [''])  # db.Predictor expects Column(Array(String))

        project = self.database_controller.get_project(name=project_name)

        if hasattr(self.handler_class, 'create_validation'):
            self.handler_class.create_validation(target, args=problem_definition)

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=target,
            learn_args=problem_definition,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            hostname=socket.gethostname(),
            version=(
                db.session.query(
                    coalesce(func.max(db.Predictor.version), 1) + (1 if is_retrain else 0)
                ).filter_by(
                    company_id=ctx.company_id,
                    name=model_name,
                    project_id=project.id,
                    deleted_at=null()
                ).scalar_subquery()),
            active=(not is_retrain),  # if create then active
        )

        db.serializable_insert(predictor_record)

        task = self.base_ml_executor.apply_async(
            task_type=ML_TASK_TYPE.LEARN,
            model_id=predictor_record.id,
            payload={
                'handler_meta': {
                    'module_path': self.handler_class.__module__,
                    'class_name': self.handler_class.__name__,
                    'engine': self.engine,
                    'integration_id': self.integration_id
                },
                'context': ctx.dump(),
                'model_id': predictor_record.id,
                'problem_definition': problem_definition,
                'set_active': set_active,
                'data_integration_ref': data_integration_ref,
                'fetch_data_query': fetch_data_query,
                'project_name': project_name
            }
        )

        if join_learn_process is True:
            task.result()
            predictor_record = db.Predictor.query.get(predictor_record.id)
            db.session.refresh(predictor_record)
        else:
            # to prevent memory leak need to add any callback
            task.add_done_callback(empty_callback)

        return predictor_record

    @profiler.profile()
    @mark_process(name='predict')
    def predict(self, model_name: str, data: list, pred_format: str = 'dict',
                project_name: str = None, version=None, params: dict = None):
        """ Generates predictions with some model and input data. """
        if isinstance(data, dict):
            data = [data]
        df = pd.DataFrame(data)
        kwargs = {
            'name': model_name,
            'ml_handler_name': self.name,
            'project_name': project_name
        }
        if version is None:
            kwargs['active'] = True
        else:
            kwargs['active'] = None
            kwargs['version'] = version
        predictor_record = get_model_record(**kwargs)
        if predictor_record is None:
            if version is not None:
                model_name = f'{model_name}.{version}'
            raise Exception(f"Error: model '{model_name}' does not exists!")
        if predictor_record.status != PREDICTOR_STATUS.COMPLETE:
            raise Exception("Error: model creation not completed")

        args = {
            'pred_format': pred_format,
            'predict_params': {} if params is None else params
        }

        try:
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.PREDICT,
                model_id=predictor_record.id,
                payload={
                    'handler_meta': {
                        'module_path': self.handler_class.__module__,
                        'class_name': self.handler_class.__name__,
                        'engine': self.engine,
                        'integration_id': self.integration_id
                    },
                    'context': ctx.dump(),
                    'predictor_record': predictor_record,
                    'args': args
                },
                dataframe=df
            )
            predictions = task.result()
        except Exception as e:
            msg = str(e).strip()
            if msg == '':
                msg = e.__class__.__name__
            msg = f'[{self.name}/{model_name}]: {msg}'
            raise MLEngineException(msg) from e

        # mdb indexes
        if '__mindsdb_row_id' not in predictions.columns and '__mindsdb_row_id' in df.columns:
            predictions['__mindsdb_row_id'] = df['__mindsdb_row_id']

        after_predict_hook(
            company_id=ctx.company_id,
            predictor_id=predictor_record.id,
            rows_in_count=df.shape[0],
            columns_in_count=df.shape[1],
            rows_out_count=len(predictions)
        )
        return predictions

    @profiler.profile()
    def finetune(
            self, model_name, project_name,
            base_model_version: int,
            data_integration_ref=None,
            fetch_data_query=None,
            join_learn_process=False,
            label=None,
            set_active=True,
            args: Optional[dict] = None
    ):
        # generate new record from latest version as starting point
        project = self.database_controller.get_project(name=project_name)

        search_args = {
            'active': None,
            'name': model_name,
            'status': PREDICTOR_STATUS.COMPLETE
        }
        if base_model_version is not None:
            search_args['version'] = base_model_version
        else:
            search_args['active'] = True
        predictor_records = get_model_records(**search_args)
        if len(predictor_records) == 0:
            raise Exception("Can't find suitable base model")

        predictor_records.sort(key=lambda x: x.training_stop_at, reverse=True)
        predictor_records = [x for x in predictor_records if x.training_stop_at is not None]
        base_predictor_record = predictor_records[0]

        learn_args = base_predictor_record.learn_args
        learn_args['using'] = args if not learn_args.get('using', False) else {**learn_args['using'], **args}

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=base_predictor_record.to_predict,
            learn_args=learn_args,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            hostname=socket.gethostname(),
            version=(
                db.session.query(
                    coalesce(func.max(db.Predictor.version), 1) + 1
                ).filter_by(
                    company_id=ctx.company_id,
                    name=model_name,
                    project_id=project.id,
                    deleted_at=null()
                ).scalar_subquery()
            ),
            active=False
        )
        db.serializable_insert(predictor_record)

        task = self.base_ml_executor.apply_async(
            task_type=ML_TASK_TYPE.FINETUNE,
            model_id=predictor_record.id,
            payload={
                'handler_meta': {
                    'module_path': self.handler_class.__module__,
                    'class_name': self.handler_class.__name__,
                    'engine': self.engine,
                    'integration_id': self.integration_id
                },
                'context': ctx.dump(),
                'model_id': predictor_record.id,
                'problem_definition': predictor_record.learn_args,
                'set_active': set_active,
                'base_model_id': base_predictor_record.id,
                'data_integration_ref': data_integration_ref,
                'fetch_data_query': fetch_data_query,
                'project_name': project_name
            }
        )

        if join_learn_process is True:
            task.result()
            predictor_record = db.Predictor.query.get(predictor_record.id)
            db.session.refresh(predictor_record)
        else:
            # to prevent memory leak need to add any callback
            task.add_done_callback(empty_callback)

        return predictor_record



from mindsdb_sql import parse_sql

from mindsdb.integrations.handlers.sharepoint_handler.sharepoint_api import (
    SharepointAPI,
)
from mindsdb.integrations.handlers.sharepoint_handler.sharepoint_tables import (
    ListsTable,
    SitesTable,
    ListItemsTable,
    SiteColumnsTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.utilities import log
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from mindsdb.utilities.log import get_log
from collections import OrderedDict


logger = get_log("integrations.github_handler")


class SharepointHandler(APIHandler):
    """
    The Sharepoint handler implementation.
    """

    name = "sharepoint"

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        if not (
            self.connection_data["clientId"]
            and (
                self.connection_data["tenantId"]
                and self.connection_data["clientSecret"]
            )
        ):
            raise Exception(
                "client params and tenant id is required for Sharepoint connection!"
            )

        self.connection = None
        self.is_connected = False
        self._client = None
        lists_data = ListsTable(self)
        self._register_table("lists", lists_data)

        sites_data = SitesTable(self)
        self._register_table("sites", sites_data)

        site_columns_data = SiteColumnsTable(self)
        self._register_table("siteColumns", site_columns_data)

        list_items_data = ListItemsTable(self)
        self._register_table("listItems", list_items_data)

    def connect(self):
        """
        Set up the context connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection
        self.connection = SharepointAPI(
            tenant_id=self.connection_data["tenantId"],
            client_id=self.connection_data["clientId"],
            client_secret=self.connection_data["clientSecret"],
        )
        self.connection.get_bearer_token()
        self.is_connected = True
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            connection = self.connect()
            response.success = connection.check_bearer_token_validity()
        except Exception as e:
            log.logger.error("Error connecting to Sharepoint! " + str(e))
            response.error_message = str(e)

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    clientId={
        "type": ARG_TYPE.STR,
        "description": "Client Id of the App",
        "required": True,
        "label": "Client ID",
    },
    clientSecret={
        "type": ARG_TYPE.PWD,
        "description": "Client Secret of the App",
        "required": True,
        "label": "Client Secret",
    },
    tenantId={
        "type": ARG_TYPE.STR,
        "description": "Tenant Id of the tenant of the App",
        "required": True,
        "label": "Tenant ID",
    },
)

connection_args_example = OrderedDict(
    clientId="xxxx-xxxx-xxxx-xxxx",
    clientSecret="<secret>",
    tenantId="xxxx-xxxx-xxxx-xxxx",
)



from copy import deepcopy

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import (
    BinaryOperation,
    Identifier,
    Constant
)

from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import SQLQuery
from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.interfaces.query_context.context_controller import query_context_controller


class ProjectDataNode(DataNode):
    type = 'project'

    def __init__(self, project, integration_controller, information_schema):
        self.project = project
        self.integration_controller = integration_controller
        self.information_schema = information_schema

    def get_type(self):
        return self.type

    def get_tables(self):
        tables = self.project.get_tables()
        table_types = {
            'table': 'BASE TABLE',
            'model': 'MODEL',
            'view': 'VIEW'
        }
        tables = [
            {
                'TABLE_NAME': key,
                'TABLE_TYPE': table_types.get(val['type'])
            }
            for key, val in tables.items()
        ]
        result = [TablesRow.from_dict(row) for row in tables]
        return result

    def has_table(self, table_name):
        tables = self.project.get_tables()
        return table_name in tables

    def get_table_columns(self, table_name):
        return self.project.get_columns(table_name)

    def predict(self, model_name: str, data, version=None, params=None):
        model_metadata = self.project.get_model(model_name)
        if model_metadata is None:
            raise Exception(f"Can't find model '{model_name}'")
        model_metadata = model_metadata['metadata']
        if model_metadata['update_status'] == 'available':
            raise Exception(f"model '{model_name}' is obsolete and needs to be updated. Run 'RETRAIN {model_name};'")
        handler = self.integration_controller.get_handler(model_metadata['engine_name'])
        return handler.predict(model_name, data, project_name=self.project.name, version=version, params=params)

    def query(self, query=None, native_query=None, session=None):
        if query is None and native_query is not None:
            query = parse_sql(native_query, dialect='mindsdb')

        # region is it query to 'models' or 'models_versions'?
        query_table = query.from_table.parts[0]
        # region FIXME temporary fix to not broke queries to 'mindsdb.models'. Can be deleted it after 1.12.2022
        if query_table == 'predictors':
            query.from_table.parts[0] = 'models'
            query_table = 'models'
        # endregion
        if query_table in ('models', 'models_versions', 'jobs', 'jobs_history', 'mdb_triggers', 'chatbots'):
            new_query = deepcopy(query)
            project_filter = BinaryOperation('=', args=[
                Identifier('project'),
                Constant(self.project.name)
            ])
            if new_query.where is None:
                new_query.where = project_filter
            else:
                new_query.where = BinaryOperation('and', args=[
                    new_query.where,
                    project_filter
                ])
            data, columns_info = self.information_schema.query(new_query)
            return data, columns_info
        # endregion

        # region query to views
        view_meta = self.project.query_view(query)

        query_context_controller.set_context('view', view_meta['id'])

        try:
            sqlquery = SQLQuery(
                view_meta['query_ast'],
                session=session
            )
            result = sqlquery.fetch(view='dataframe')

        finally:
            query_context_controller.release_context('view', view_meta['id'])

        if result['success'] is False:
            raise Exception(f"Cant execute view query: {view_meta['query_ast']}")
        df = result['result']

        df = query_df(df, query)

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in df.dtypes.items()
        ]

        return df.to_dict(orient='records'), columns_info
        # endregion



import json
import requests
from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.llm_utils import get_completed_prompts


class OllamaHandler(BaseMLEngine):
    name = "ollama"
    SERVE_URL = 'http://localhost:11434'
    MODEL_LIST_URL = 'https://registry.ollama.ai/v2/_catalog'

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'using' not in args:
            raise Exception("Ollama engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args['using']

        # check model version is valid
        try:
            all_models = requests.get(OllamaHandler.MODEL_LIST_URL).json()['repositories']
        except Exception as e:
            raise Exception(f"Could not retrieve model list from Ollama registry: {e}")
        base_models = list(filter(lambda x: 'library/' in x, all_models))
        valid_models = [m.split('/')[-1] for m in base_models]

        if 'model_name' not in args:
            raise Exception('`model_name` must be provided in the USING clause.')
        elif args['model_name'] not in valid_models:
            raise Exception(f"The model `{args['model_name']}` is not yet supported by Ollama! Please choose one of the following: {valid_models}")  # noqa

        # check ollama service health
        status = requests.get(OllamaHandler.SERVE_URL + '/api/tags').status_code
        if status != 200:
            raise Exception(f"Ollama service is not working (status `{status}`). Please double check it is running and try again.")  # noqa

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """ Pull LLM artifacts with Ollama API. """
        # arg setter
        args = args['using']
        args['target'] = target
        self.model_storage.json_set('args', args)

        # download model
        # TODO v2: point Ollama to the engine storage folder instead of their default location
        model_name = args['model_name']
        # blocking operation, finishes once model has been fully pulled and served
        requests.post(OllamaHandler.SERVE_URL + '/api/pull', json={'name': model_name})

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        """
        Generate text completions with the local LLM.
            Args:
                df (pd.DataFrame): The input DataFrame containing data to predict.
                args (Optional[Dict]): Additional arguments for prediction parameters.
            Returns:
                pd.DataFrame: The DataFrame containing row-wise text completions.
        """
        # setup
        pred_args = args.get('predict_params', {})
        args = self.model_storage.json_get('args')
        model_name, target_col = args['model_name'], args['target']
        prompt_template = pred_args.get('prompt_template',
                                        args.get('prompt_template', 'Answer the following question: {{{{text}}}}'))
        # TODO v2: add support for overriding modelfile params (e.g. temperature)

        # prepare prompts
        prompts, empty_prompt_ids = get_completed_prompts(prompt_template, df)
        df['__mdb_prompt'] = prompts

        # call llm
        completions = []
        for i, row in df.iterrows():
            if i not in empty_prompt_ids:
                raw_output = requests.post(
                    OllamaHandler.SERVE_URL + '/api/generate',
                    json={
                        'model': args['model_name'],
                        'prompt': row['__mdb_prompt'],
                    }
                )
                out_tokens = raw_output.content.decode().split('\n')  # stream of output tokens

                tokens = []
                for o in out_tokens:
                    if o != '':
                        # TODO v2: add support for storing `context` short conversational memory
                        info = json.loads(o)
                        if 'response' in info:
                            token = info['response']
                            tokens.append(token)

                completions.append(''.join(tokens))
            else:
                completions.append('')

        # consolidate output
        data = pd.DataFrame(completions)
        data.columns = [target_col]
        return data

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        model_name, target_col = args['model_name'], args['target']
        prompt_template = args.get('prompt_template', 'Answer the following question: {{{{text}}}}')

        if attribute == "features":
            return pd.DataFrame([[target_col, prompt_template]], columns=['target_column', 'mindsdb_prompt_template'])

        # get model info
        else:
            model_info = requests.post(OllamaHandler.SERVE_URL + '/api/show', json={'name': model_name}).json()
            return pd.DataFrame([[
                model_name,
                model_info['license'],
                model_info['modelfile'],
                model_info['parameters'],
                model_info['template'],
            ]],
                columns=[
                    'model_type',
                    'license',
                    'modelfile',
                    'parameters',
                    'ollama_base_template',
                ])



from ..mysql_handler import Handler as MySQLHandler


class ApacheDorisHandler(MySQLHandler):
    """This handler handles connection and execution of the Apache Doris statements."""

    name = 'apache_doris'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



import unittest
from mindsdb.integrations.handlers.ignite_handler.ignite_handler import IgniteHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class IgniteHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "127.0.0.1",
            "port": 10800
        }
        cls.handler = IgniteHandler('test_ignite_handler', cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_native_query_select(self):
        query = "SELECT * FROM City"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_4_get_columns(self):
        columns = self.handler.get_columns('City')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



import base64
import secrets
import urllib
import time

import requests
from flask import request, session, redirect, url_for
from flask_restx import Resource


from mindsdb.api.http.namespaces.configs.auth import ns_conf
from mindsdb.utilities.config import Config
from mindsdb.utilities.log import get_log


log = get_log('http')


def get_access_token() -> str:
    ''' return current access token

        Returns:
            str: token
    '''
    return Config().get('auth', {}).get('oauth', {}).get('tokens', {}).get('access_token')


def request_user_info(access_token: str = None) -> dict:
    ''' request user info from cloud

        Args:
            access_token (str, optional): token that used to get user data

        Returns:
            dict: user data
    '''
    if access_token is None:
        access_token = get_access_token()
    if access_token is None:
        raise KeyError()

    auth_server = Config()['auth']['oauth']['server_host']

    response = requests.get(
        f'https://{auth_server}/auth/userinfo',
        headers={
            'Authorization': f'Bearer {access_token}'
        },
        timeout=5
    )
    if response.status_code != 200:
        raise Exception(f'Wrong response: {response.status_code}, {response.text}')

    return response.json()


@ns_conf.route('/callback', methods=['GET'])
@ns_conf.route('/callback/cloud_home', methods=['GET'])
@ns_conf.hide
class Auth(Resource):
    @ns_conf.doc(
        params={'code': 'authentification code'}
    )
    def get(self):
        ''' callback from auth server if authentification is successful
        '''
        config = Config()
        code = request.args.get('code')

        aws_meta_data = config['aws_meta_data']
        public_hostname = aws_meta_data['public-hostname']
        instance_id = aws_meta_data['instance-id']

        oauth_meta = config['auth']['oauth']
        client_id = oauth_meta['client_id']
        client_secret = oauth_meta['client_secret']
        auth_server = oauth_meta['server_host']
        client_basic = base64.b64encode(
            f'{client_id}:{client_secret}'.encode()
        ).decode()

        redirect_uri = f'https://{public_hostname}{request.path}'
        response = requests.post(
            f'https://{auth_server}/auth/token',
            data={
                'code': code,
                'grant_type': 'authorization_code',
                'redirect_uri': redirect_uri
            },
            headers={
                'Authorization': f'Basic {client_basic}'
            }
        )
        tokens = response.json()
        if 'expires_in' in tokens:
            tokens['expires_at'] = round(time.time() + tokens['expires_in'] - 1)
            del tokens['expires_in']

        user_data = request_user_info(tokens['access_token'])

        previous_username = config['auth']['oauth'].get('username')
        new_username = user_data['name']
        if (
            previous_username is not None
            and new_username != previous_username
        ):
            return redirect('/forbidden')

        config.update({
            'auth': {
                'provider': 'cloud',
                'oauth': {
                    'username': new_username,
                    'tokens': tokens
                }
            }
        })

        try:
            resp = requests.put(
                f'https://{auth_server}/cloud/instance',
                json={
                    'instance_id': instance_id,
                    'public_hostname': public_hostname,
                    'ami_id': aws_meta_data.get('ami-id')
                },
                headers={
                    'Authorization': f'Bearer {tokens["access_token"]}'
                },
                timeout=5
            )
            if resp.status_code != 200:
                log.warn(f'Wrong response from cloud server: {resp.status_code}')
        except Exception as e:
            log.warn(f"Cant't send request to cloud server: {e}")

        session['username'] = user_data['name']
        session['auth_provider'] = 'cloud'
        session.permanent = True

        if request.path.endswith('/auth/callback/cloud_home'):
            return redirect(f'https://{auth_server}')
        else:
            return redirect(url_for('root_index'))


@ns_conf.route('/cloud_login', methods=['GET'])
@ns_conf.hide
class CloudLoginRoute(Resource):
    @ns_conf.doc(
        responses={
            302: 'Redirect to auth server'
        },
        params={'location': 'final redirection should lead to that location'}
    )
    def get(self):
        ''' redirect to cloud login form
        '''
        location = request.args.get('location')
        config = Config()

        aws_meta_data = config['aws_meta_data']
        public_hostname = aws_meta_data['public-hostname']
        auth_server = config['auth']['oauth']['server_host']

        if location == 'cloud_home':
            redirect_uri = f'https://{public_hostname}/api/auth/callback/cloud_home'
        else:
            redirect_uri = f'https://{public_hostname}/api/auth/callback'

        args = urllib.parse.urlencode({
            'client_id': config['auth']['oauth']['client_id'],
            'scope': 'openid profile aws_marketplace',
            'response_type': 'code',
            'nonce': secrets.token_urlsafe(),
            'redirect_uri': redirect_uri
        })
        return redirect(f'https://{auth_server}/auth/authorize?{args}')



from mindsdb_sql import parse_sql

from mindsdb.integrations.handlers.npm_handler.api import NPM
from mindsdb.integrations.handlers.npm_handler.npm_tables import (
    NPMMetadataTable,
    NPMMaintainersTable,
    NPMKeywordsTable,
    NPMDependenciesTable,
    NPMDevDependenciesTable,
    NPMOptionalDependenciesTable,
    NPMGithubStatsTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse


class NPMHandler(APIHandler):

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(name)
        self.connection = None
        self.is_connected = False
        _tables = [
            NPMMetadataTable,
            NPMMaintainersTable,
            NPMKeywordsTable,
            NPMDependenciesTable,
            NPMDevDependenciesTable,
            NPMOptionalDependenciesTable,
            NPMGithubStatsTable,
        ]
        for Table in _tables:
            self._register_table(Table.name, Table(self))

    def check_connection(self) -> StatusResponse:
        """Check if connected"""
        response = StatusResponse(False)
        if NPM.is_connected():
            response.success = True
        else:
            response.success = False
        self.is_connected = True
        return response

    def connect(self) -> NPM:
        """Make connection object"""
        self.connection = NPM
        return self.connection

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query"""
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)



from collections import OrderedDict

from mindsdb.integrations.handlers.oilpriceapi_handler.oilpriceapi_tables import (
    OilPriceLatestTable,
    OilPricePastDayPriceTable
)
from mindsdb.integrations.handlers.oilpriceapi_handler.oilpriceapi import OilPriceAPIClient
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from mindsdb.utilities.log import get_log
from mindsdb_sql import parse_sql


logger = get_log("integrations.oilpriceapi_handler")


class OilPriceAPIHandler(APIHandler):
    """The OilPriceAPI handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the OilPriceAPI handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.client = OilPriceAPIClient(self.connection_data["api_key"])
        self.is_connected = False

        latest_price_data = OilPriceLatestTable(self)
        self._register_table("latest_price", latest_price_data)

        past_day_price_data = OilPricePastDayPriceTable(self)
        self._register_table("past_day_price", past_day_price_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """
        resp = StatusResponse(False)
        status = self.client.get_latest_price()
        if status["code"] != 200:
            resp.success = False
            resp.error_message = status["error"]
            return resp
        self.is_connected = True
        return resp

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            status = self.client.get_latest_price()
            if status["code"] == 200:
                logger.info("Authentication successful")
                response.success = True
            else:
                response.success = False
                logger.info("Error connecting to OilPriceAPI. " + status["error"])
                response.error_message = status["error"]
        except Exception as e:
            logger.error(f"Error connecting to OilPriceAPI: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.PWD,
        "description": "OilPriceAPI key to use for authentication.",
        "required": True,
        "label": "Api key",
    }
)

connection_args_example = OrderedDict(
    api_key=""
)



from mindsdb.utilities.config import Config
from mindsdb.api.mongo.server import run_server
from mindsdb.interfaces.storage import db
from mindsdb.utilities.log import initialize_log
from mindsdb.utilities.functions import init_lexer_parsers


def start(verbose=False):
    config = Config()
    db.init()
    init_lexer_parsers()

    initialize_log(config, 'mongodb', wrap_print=True)

    run_server(config)



import requests
from requests import Response
from enum import Enum


class TripAdvisorAPICall(Enum):
    """TripAdvisor API references"""

    SEARCH_LOCATION = 1
    LOCATION_DETAILS = 2
    PHOTOS = 3
    REVIEWS = 4
    NEARBY_SEARCH = 5


class TripAdvisorAPI:
    """A class for checking the connection to the TripAdvisor Content API and making requests.

    Attributes:
        api_key (str): The unique API key to access Tripadvisor content.
    """

    def __init__(self, api_key):
        self.api_key = api_key

    def checkTripAdvisorConnection(self):
        """
        Check the connection with TripAdvisor
        """
        url = "https://api.content.tripadvisor.com/api/v1/location/search?language=en&key={api_key}&searchQuery={searchQuery}".format(
            api_key=self.api_key, searchQuery="London"
        )

        headers = {"accept": "application/json"}
        response = requests.get(url, headers=headers)
        status_code = response.status_code

        if status_code >= 400 and status_code <= 499:
            raise Exception("Client error: " + response.text)

        if status_code >= 500 and status_code <= 599:
            raise Exception("Server error: " + response.text)

    def getResponse(self, url: str) -> Response:
        """
        Getting a response from the API call
        """
        headers = {"accept": "application/json"}
        response = requests.get(url, headers=headers)
        return response

    def getURLQuery(self, url: str, params_dict: dict) -> str:
        """
        Processing the query and adding parameters to the URL
        """
        for idx, (queryParam, value) in enumerate(params_dict.items()):
            if value is not None or value != "":
                if value != "" and any(
                    next_value != "" or value is not None
                    for next_value in list(params_dict.values())[idx + 1:]
                ):
                    url += "{queryParam}={value}&".format(
                        queryParam=queryParam, value=value
                    )
                else:
                    url += "{queryParam}={value}".format(
                        queryParam=queryParam, value=value
                    )
        return url

    def location_search(
        self,
        url: str,
        params_dict: dict,
        language: str = "en",
    ) -> Response:
        """
        The Location Search request returns up to 10 locations found by the given search query. You can use category ("hotels", "attractions", "restaurants", "geos"),
        phone number, address, and latitude/longtitude to search with more accuracy.

        Args:
            searchQuery (str): Text to use for searching based on the name of the location.
            category (str): Filters result set based on property type. Valid options are "hotels", "attractions", "restaurants", and "geos".
            phone (str): Phone number to filter the search results by (this can be in any format with spaces and dashes but without the "+" sign at the beginning).
            address (str): Address to filter the search results by.
            latLong (str): Latitude/Longitude pair to scope down the search around a specifc point - eg. "42.3455,-71.10767".
            radius (int): Length of the radius from the provided latitude/longitude pair to filter results.
            radiusUnit (str): Unit for length of the radius. Valid options are "km", "mi", "m" (km=kilometers, mi=miles, m=meters.
            language (str): The language in which to return results (e.g. "en" for English or "es" for Spanish) from the list of our Supported Languages.

        Returns:
            response: Response object with response data as application/json
        """

        url = url + "search?language={language}&key={api_key}&".format(
            api_key=self.api_key, language=language
        )

        url = self.getURLQuery(url, params_dict)
        response = self.getResponse(url)

        return response

    def location_details(
        self, url: str, params_dict: dict, locationId: str, language: str = "en"
    ) -> Response:
        """
        A Location Details request returns comprehensive information about a location (hotel, restaurant, or an attraction) such as name, address, rating, and URLs for the listing
        on Tripadvisor.

        Args:
            locationId (str): A unique identifier for a location on Tripadvisor. The location ID can be obtained using the Location Search.
            language (str): The language in which to return results (e.g. "en" for English or "es" for Spanish) from the list of our Supported Languages.
            currency (str): The currency code to use for request and response (should follow ISO 4217).

        Returns:
            response (Response): Response object with response data as application/json
        """
        url = url + "{locationId}/details?language={language}&key={api_key}&".format(
            locationId=locationId, api_key=self.api_key, language=language
        )
        url = self.getURLQuery(url, params_dict)
        response = self.getResponse(url)
        return response

    def location_reviews(
        self, url: str, locationId: str, language: str = "en"
    ) -> Response:
        """
        The Location Reviews request returns up to 5 of the most recent reviews for a specific location. Please note that the limits are different for the beta subscribers.

        Args:
            locationId (str): A unique identifier for a location on Tripadvisor. The location ID can be obtained using the Location Search.
            language (str): The language in which to return results (e.g. "en" for English or "es" for Spanish) from the list of our Supported Languages.

        Returns:
            response: Response object with response data as application/json
        """
        url = url + "{locationId}/reviews?language={language}&key={api_key}&".format(
            locationId=locationId, api_key=self.api_key, language=language
        )

        response = self.getResponse(url)
        return response

    def location_photos(self, url: str, locationId: str, language: str = "en") -> Response:
        """
        The Location Photos request returns up to 5 high-quality photos for a specific location. Please note that the limits are different for the beta subscribers.
        You need to upgrade to get the higher limits mentioned here. The photos are ordered by recency.

        Args:
            locationId (str): A unique identifier for a location on Tripadvisor. The location ID can be obtained using the Location Search.
            language (str): The language in which to return results (e.g. "en" for English or "es" for Spanish) from the list of our Supported Languages.

        Returns:
            response: Response object with response data as application/json
        """
        url = url + "{locationId}/photos?language={language}&key={api_key}".format(locationId=locationId, language=language, api_key=self.api_key)
        response = self.getResponse(url)
        return response

    def location_nearby_search(self, url: str, params_dict: dict, language: str = "en") -> Response:
        """
        The Nearby Location Search request returns up to 10 locations found near the given latitude/longtitude.
        You can use category ("hotels", "attractions", "restaurants", "geos"), phone number, address to search with more accuracy.

        Args:
            latLong (str): Latitude/Longitude pair to scope down the search around a specifc point - eg. "42.3455,-71.10767".
            category (str): Filters result set based on property type. Valid options are "hotels", "attractions", "restaurants", and "geos".
            phone (str): Phone number to filter the search results by (this can be in any format with spaces and dashes but without the "+" sign at the beginning).
            address (str): Address to filter the search results by.
            radius (int): Length of the radius from the provided latitude/longitude pair to filter results.
            radiusUnit (str): Unit for length of the radius. Valid options are "km", "mi", "m" (km=kilometers, mi=miles, m=meters.
            language (str): The language in which to return results (e.g. "en" for English or "es" for Spanish) from the list of our Supported Languages.

        Returns:
            response: Response object with response data as application/json
        """

        url = url + "nearby_search?language={language}&key={api_key}&".format(
            api_key=self.api_key, language=language
        )

        url = self.getURLQuery(url, params_dict)
        response = self.getResponse(url)
        return response

    def getTripAdvisorData(self, apiCall, **params):
        """
        Making a request based on the query and receive data from TripAdvisor.
        """
        url = "https://api.content.tripadvisor.com/api/v1/location/"
        params_dict = params

        if apiCall == TripAdvisorAPICall.SEARCH_LOCATION:
            response = self.location_search(url, params_dict)
            return response.json()["data"]

        elif apiCall == TripAdvisorAPICall.LOCATION_DETAILS:
            response = self.location_details(
                url, params_dict, params_dict["locationId"]
            )
            return response.json()

        elif apiCall == TripAdvisorAPICall.REVIEWS:
            response = self.location_reviews(url, params_dict["locationId"])
            return response.json()["data"]

        elif apiCall == TripAdvisorAPICall.PHOTOS:
            response = self.location_photos(url, params_dict["locationId"])
            return response.json()["data"]

        elif apiCall == TripAdvisorAPICall.NEARBY_SEARCH:
            response = self.location_nearby_search(url, params_dict)
            return response.json()["data"]



from typing import Optional
from collections import OrderedDict

import pandas as pd
import fdb

from mindsdb_sql import parse_sql
from sqlalchemy_firebird.base import FBDialect
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class FirebirdHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Firebird statements.
    """

    name = 'firebird'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'firebird'
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = fdb.connect(
            host=self.connection_data['host'],
            database=self.connection_data['database'],
            user=self.connection_data['user'],
            password=self.connection_data['password'],
        )
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to Firebird {self.connection_data["database"]}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        cursor = connection.cursor()

        try:
            cursor.execute(query)
            result = cursor.fetchall()
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result,
                        columns=[x[0] for x in cursor.description]
                    )
                )
            else:
                connection.commit()
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            log.logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        cursor.close()
        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender(FBDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = """
            SELECT RDB$RELATION_NAME 
            FROM RDB$RELATIONS
            WHERE (RDB$SYSTEM_FLAG <> 1 OR RDB$SYSTEM_FLAG IS NULL) AND RDB$VIEW_BLR IS NULL
            ORDER BY RDB$RELATION_NAME;
        """
        result = self.native_query(query)
        df = result.data_frame
        df[df.columns[0]] = df[df.columns[0]].apply(lambda row: row.strip())
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"""
            SELECT
              RF.RDB$FIELD_NAME FIELD_NAME,
              CASE F.RDB$FIELD_TYPE
                WHEN 7 THEN
                  CASE F.RDB$FIELD_SUB_TYPE
                    WHEN 0 THEN 'SMALLINT'
                    WHEN 1 THEN 'NUMERIC(' || F.RDB$FIELD_PRECISION || ', ' || (-F.RDB$FIELD_SCALE) || ')'
                    WHEN 2 THEN 'DECIMAL'
                  END
                WHEN 8 THEN
                  CASE F.RDB$FIELD_SUB_TYPE
                    WHEN 0 THEN 'INTEGER'
                    WHEN 1 THEN 'NUMERIC('  || F.RDB$FIELD_PRECISION || ', ' || (-F.RDB$FIELD_SCALE) || ')'
                    WHEN 2 THEN 'DECIMAL'
                  END
                WHEN 9 THEN 'QUAD'
                WHEN 10 THEN 'FLOAT'
                WHEN 12 THEN 'DATE'
                WHEN 13 THEN 'TIME'
                WHEN 14 THEN 'CHAR(' || (TRUNC(F.RDB$FIELD_LENGTH / CH.RDB$BYTES_PER_CHARACTER)) || ') '
                WHEN 16 THEN
                  CASE F.RDB$FIELD_SUB_TYPE
                    WHEN 0 THEN 'BIGINT'
                    WHEN 1 THEN 'NUMERIC(' || F.RDB$FIELD_PRECISION || ', ' || (-F.RDB$FIELD_SCALE) || ')'
                    WHEN 2 THEN 'DECIMAL'
                  END
                WHEN 27 THEN 'DOUBLE'
                WHEN 35 THEN 'TIMESTAMP'
                WHEN 37 THEN 'VARCHAR(' || (TRUNC(F.RDB$FIELD_LENGTH / CH.RDB$BYTES_PER_CHARACTER)) || ')'
                WHEN 40 THEN 'CSTRING' || (TRUNC(F.RDB$FIELD_LENGTH / CH.RDB$BYTES_PER_CHARACTER)) || ')'
                WHEN 45 THEN 'BLOB_ID'
                WHEN 261 THEN 'BLOB SUB_TYPE ' || F.RDB$FIELD_SUB_TYPE
                ELSE 'RDB$FIELD_TYPE: ' || F.RDB$FIELD_TYPE || '?'
              END FIELD_TYPE,
              IIF(COALESCE(RF.RDB$NULL_FLAG, 0) = 0, NULL, 'NOT NULL') FIELD_NULL,
              CH.RDB$CHARACTER_SET_NAME FIELD_CHARSET,
              DCO.RDB$COLLATION_NAME FIELD_COLLATION,
              COALESCE(RF.RDB$DEFAULT_SOURCE, F.RDB$DEFAULT_SOURCE) FIELD_DEFAULT,
              F.RDB$VALIDATION_SOURCE FIELD_CHECK,
              RF.RDB$DESCRIPTION FIELD_DESCRIPTION
            FROM RDB$RELATION_FIELDS RF
            JOIN RDB$FIELDS F ON (F.RDB$FIELD_NAME = RF.RDB$FIELD_SOURCE)
            LEFT OUTER JOIN RDB$CHARACTER_SETS CH ON (CH.RDB$CHARACTER_SET_ID = F.RDB$CHARACTER_SET_ID)
            LEFT OUTER JOIN RDB$COLLATIONS DCO ON ((DCO.RDB$COLLATION_ID = F.RDB$COLLATION_ID) AND (DCO.RDB$CHARACTER_SET_ID = F.RDB$CHARACTER_SET_ID))
            WHERE (RF.RDB$RELATION_NAME = '{table_name.upper()}') AND (COALESCE(RF.RDB$SYSTEM_FLAG, 0) = 0)
            ORDER BY RF.RDB$FIELD_POSITION;
        """
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={'FIELD_NAME': 'column_name', 'FIELD_TYPE': 'data_type'})
        return result


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Firebird server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': """
            The database name to use when connecting with the Firebird server. NOTE: use double backslashes (\\) for the 
            database path on a Windows machine.
        """
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Firebird server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Firebird server.'
    }
)

connection_args_example = OrderedDict(
    host='localhost',
    database='/temp/test.db',
    user='sysdba',
    password='password'
)


import unittest
from mindsdb.integrations.handlers.clickhouse_handler.clickhouse_handler import ClickHouseHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class PostgresHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        connection_data = {
            "host": "localhost",
            "port": "9000",
            "user": "root",
            "password": "pass",
            "database": "test_data"
        }
        cls.handler = ClickHouseHandler('test_clickhouse_handler', connection_data)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_native_query_show_dbs(self):
        result = self.handler.native_query("SHOW DATABASES;")
        assert result.type is not RESPONSE_TYPE.ERROR

    def test_2_wrong_native_query_returns_error(self):
        result = self.handler.native_query("SHOW DATABASE1S;")
        assert result.type is RESPONSE_TYPE.ERROR

    def test_3_select_query(self):
        query = 'SELECT * FROM hdi'
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls.type is not RESPONSE_TYPE.ERROR

    def test_5_describe_table(self):
        described = self.handler.get_columns("hdi")
        print('described', described)
        assert described.type is RESPONSE_TYPE.TABLE



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'replSetGetStatus': helpers.is_true}

    result = {
        "ok": 0
    }


responder = Responce()



import traceback
import threading
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log

from mindsdb.interfaces.triggers.trigger_task import TriggerTask
from mindsdb.interfaces.chatbot.chatbot_task import ChatBotTask


class TaskThread(threading.Thread):

    def __init__(self, task_id):
        threading.Thread.__init__(self)
        self.task_id = task_id
        self._stop_event = threading.Event()
        self.object_type = None
        self.object_id = None

    def run(self):
        # create context and session

        task_record = db.Tasks.query.get(self.task_id)

        ctx.set_default()
        ctx.company_id = task_record.company_id
        if task_record.user_class is not None:
            ctx.user_class = task_record.user_class

        self.object_type = task_record.object_type
        self.object_id = task_record.object_id

        log.logger.info(f'Task starting: {self.object_type}.{self.object_id}')
        try:
            if self.object_type == 'trigger':

                trigger = TriggerTask(self.task_id, self.object_id)
                trigger.run(self._stop_event)

            elif self.object_type == 'chatbot':
                bot = ChatBotTask(self.task_id, self.object_id)
                bot.run(self._stop_event)

        except Exception:
            log.logger.error(traceback.format_exc())
            task_record.last_error = str(traceback.format_exc())

        db.session.commit()

    def stop(self):
        log.logger.info(f'Task stopping: {self.object_type}.{self.object_id}')

        self._stop_event.set()



import os
import re
import sys
import shutil
import pickle
import subprocess
import traceback
from enum import Enum
from pathlib import Path
from datetime import datetime
from typing import Optional, Dict
from collections import OrderedDict

import numpy as np
import pandas as pd
from pandas.api import types as pd_types

from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage import db
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.integrations.utilities.utils import format_exception_error
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
import mindsdb.utilities.profiler as profiler

from .proc_wrapper import pd_decode, pd_encode, encode, decode
from .proc_wrapper import import_string, find_model_class
from .const import BYOM_METHOD


BYOM_TYPE = Enum('BYOM_TYPE', ['SAFE', 'UNSAFE'])


class BYOMHandler(BaseMLEngine):

    name = 'byom'

    def __init__(self, model_storage, engine_storage, **kwargs) -> None:
        # region check availability
        is_cloud = Config().get('cloud', False)
        if is_cloud is True:
            byom_enabled = os.environ.get('MINDSDB_BYOM_ENABLED', 'false').lower()
            if byom_enabled not in ('true', '1'):
                raise RuntimeError('BYOM is disabled on cloud')
        # endregion

        self.model_wrapper = None

        # region read MINDSDB_BYOM_TYPE
        try:
            self._byom_type = BYOM_TYPE[
                os.environ.get(
                    'MINDSDB_BYOM_TYPE',
                    BYOM_TYPE.UNSAFE.name
                ).upper()
            ]
        except KeyError:
            self._byom_type = BYOM_TYPE.SAFE
        # endregion

        super().__init__(model_storage, engine_storage, **kwargs)

    def _get_model_proxy(self):
        self.engine_storage.fileStorage.pull()
        code = self.engine_storage.fileStorage.file_get('code')
        modules_str = self.engine_storage.fileStorage.file_get('modules')

        if self.model_wrapper is None:
            if self._byom_type == BYOM_TYPE.UNSAFE:
                WrapperClass = ModelWrapperUnsafe
            elif self._byom_type == BYOM_TYPE.SAFE:
                WrapperClass = ModelWrapperSafe

            self.model_wrapper = WrapperClass(
                code=code,
                modules_str=modules_str,
                engine_id=self.engine_storage.integration_id
            )

        return self.model_wrapper

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        mp = self._get_model_proxy()
        model_state = self.model_storage.file_get('model')
        return mp.describe(model_state, attribute)

    def create(self, target, df=None, args=None, **kwargs):
        model_proxy = self._get_model_proxy()

        model_state = model_proxy.train(df, target, args)

        self.model_storage.file_set('model', model_state)

        # TODO return columns?

        def convert_type(field_type):
            if pd_types.is_integer_dtype(field_type):
                return 'integer'
            elif pd_types.is_numeric_dtype(field_type):
                return 'float'
            elif pd_types.is_datetime64_any_dtype(field_type):
                return 'datetime'
            else:
                return 'categorical'

        columns = {
            target: convert_type(object)
        }

        self.model_storage.columns_set(columns)

    def predict(self, df, args=None):
        pred_args = args.get('predict_params', {})

        model_proxy = self._get_model_proxy()

        model_state = self.model_storage.file_get('model')

        pred_df = model_proxy.predict(df, model_state, pred_args)

        return pred_df

    def create_engine(self, connection_args):
        self.engine_storage.fileStorage.file_set(
            'code',
            Path(connection_args['code']).read_bytes()
        )

        self.engine_storage.fileStorage.file_set(
            'modules',
            Path(connection_args['modules']).read_bytes()
        )

        self.engine_storage.fileStorage.push()

        model_proxy = self._get_model_proxy()

        try:
            model_proxy.check()
        except Exception as e:
            model_proxy.remove_venv()
            raise e

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        model_storage = self.model_storage

        # TODO: should probably refactor at some point, as a bit of the logic is shared with lightwood's finetune logic
        try:
            base_predictor_id = args['base_model_id']
            base_predictor_record = db.Predictor.query.get(base_predictor_id)
            if base_predictor_record.status != PREDICTOR_STATUS.COMPLETE:
                raise Exception("Base model must be in status 'complete'")

            predictor_id = model_storage.predictor_id
            predictor_record = db.Predictor.query.get(predictor_id)

            predictor_record.data = {'training_log': 'training'} # TODO move to ModelStorage (don't work w/ db directly)
            predictor_record.training_start_at = datetime.now()
            predictor_record.status = PREDICTOR_STATUS.FINETUNING  # TODO: parallel execution block
            db.session.commit()

            model_proxy = self._get_model_proxy()
            model_state = self.base_model_storage.file_get('model')
            model_state = model_proxy.finetune(df, model_state, args=args.get('using', {}))  # WRONG

            # region hack to speedup file saving
            with profiler.Context('finetune-byom-write-file'):
                dest_abs_path = self.model_storage.fileStorage.folder_path / 'model'
                with open(dest_abs_path, 'wb') as fd:
                    fd.write(model_state)
                self.model_storage.fileStorage.push(compression_level=0)
            # endregion

            predictor_record.update_status = 'up_to_date'
            predictor_record.status = PREDICTOR_STATUS.COMPLETE
            predictor_record.training_stop_at = datetime.now()
            db.session.commit()

        except Exception as e:
            log.logger.error(e)
            predictor_id = model_storage.predictor_id
            predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
            print(traceback.format_exc())
            error_message = format_exception_error(e)
            predictor_record.data = {"error": error_message}
            predictor_record.status = PREDICTOR_STATUS.ERROR
            db.session.commit()
            raise

        finally:
            if predictor_record.training_stop_at is None:
                predictor_record.training_stop_at = datetime.now()
                db.session.commit()


class ModelWrapperUnsafe:
    """ Model wrapper that executes learn/predict in current process
    """

    def __init__(self, code, modules_str, engine_id):
        module = import_string(code)
        model_class = find_model_class(module)
        self.model_class = model_class
        self.model_instance = self.model_class()

    def train(self, df, target, args):
        self.model_instance.train(df, target, args)
        return pickle.dumps(self.model_instance.__dict__, protocol=5)

    def predict(self, df, model_state, args):
        model_state = pickle.loads(model_state)
        self.model_instance.__dict__ = model_state
        try:
            result = self.model_instance.predict(df, args)
        except:
            result = self.model_instance.predict(df)
        return result

    def finetune(self, df, model_state, args):
        self.model_instance.__dict__ = pickle.loads(model_state)

        call_args = [df]
        if args:
            call_args.append(args)

        self.model_instance.finetune(df, args)

        return pickle.dumps(self.model_instance.__dict__, protocol=5)

    def describe(self, model_state, attribute: Optional[str] = None) -> pd.DataFrame:
        if hasattr(self.model_instance, 'describe'):
            model_state = pickle.loads(model_state)
            self.model_instance.__dict__ = model_state
            return self.model_instance.describe(attribute)
        return pd.DataFrame()

    def check(self):
        pass


class ModelWrapperSafe:
    """ Model wrapper that executes learn/predict in venv
    """

    def __init__(self, code, modules_str, engine_id):
        self.code = code
        modules = self.parse_requirements(modules_str)

        self.env_path = None
        self.prepare_env(modules, engine_id)

    def prepare_env(self, modules, engine_id):
        config = Config()

        try:
            import virtualenv

            base_path = config.get('byom', {}).get('venv_path')
            if base_path is None:
                # create in root path
                base_path = Path(config.paths['root']) / 'venvs'

            self.env_path = base_path / f'env_{engine_id}'

            self.python_path = self.env_path / 'bin' / 'python'

            if self.env_path.exists():
                # already exists. it means requirements are already installed
                return

            # create
            virtualenv.cli_run(['-p', sys.executable, str(self.env_path)])
            log.logger.info(f"Created new environment: {self.env_path}")

            if len(modules) > 0:
                self.install_modules(modules)

        except Exception as e:
            # DANGER !!! VENV MUST BE CREATED
            log.logger.info("Can't create virtual environment. venv module should be installed")

            self.python_path = Path(sys.executable)

            # try to install modules everytime
            self.install_modules(modules)

    def remove_venv(self):
        if self.env_path is not None and self.env_path.exists():
            shutil.rmtree(str(self.env_path))

    def parse_requirements(self, requirements):
        # get requirements from string
        # they should be located at the top of the file, before code

        pattern = '^[\w\\[\\]-]+[=!<>\s]*[\d\.]*[,=!<>\s]*[\d\.]*$'
        modules = []
        for line in requirements.split(b'\n'):
            line = line.decode().strip()
            if line:
                if re.match(pattern, line):
                    modules.append(line)
                else:
                    raise Exception(f'Wrong requirement: {line}')

        is_pandas = any([m.lower().startswith('pandas') for m in modules])
        if not is_pandas:
            modules.append('pandas >=2.0.0, <2.1.0')

        # for dataframe serialization
        modules.append('pyarrow==11.0.0')
        return modules

    def install_modules(self, modules):
        # install in current environment using pip

        pip_cmd = self.python_path.parent / 'pip'
        for module in modules:
            p = subprocess.Popen([pip_cmd, 'install', module], stderr=subprocess.PIPE)
            p.wait()
            if p.returncode != 0:
                raise Exception(f'Problem with installing module {module}: {p.stderr.read()}')

    def _run_command(self, params):
        params_enc = encode(params)

        wrapper_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'proc_wrapper.py')
        p = subprocess.Popen(
            [str(self.python_path), wrapper_path],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        p.stdin.write(params_enc)
        p.stdin.close()
        ret_enc = p.stdout.read()

        p.wait()

        try:
            ret = decode(ret_enc)
        except (pickle.UnpicklingError, EOFError):
            raise RuntimeError(p.stderr.read())
        return ret

    def check(self):

        params = {
            'method': BYOM_METHOD.CHECK,
            'code': self.code,
        }
        return self._run_command(params)

    def train(self, df, target, args):
        params = {
            'method': BYOM_METHOD.TRAIN,
            'code': self.code,
            'df': pd_encode(df),
            'to_predict': target,
            'args': args,
        }

        model_state = self._run_command(params)
        return model_state

    def predict(self, df, model_state, args):

        params = {
            'method': BYOM_METHOD.PREDICT,
            'code': self.code,
            'model_state': model_state,
            'df': pd_encode(df),
            'args': args,
        }
        pred_df = self._run_command(params)
        return pd_decode(pred_df)

    def finetune(self, df, model_state, args):
        params = {
            'method': BYOM_METHOD.FINETUNE,
            'code': self.code,
            'model_state': model_state,
            'df': pd_encode(df),
            'args': args,
        }

        model_state = self._run_command(params)
        return model_state

    def describe(self, model_state, attribute: Optional[str] = None) -> pd.DataFrame:
        params = {
            'method': BYOM_METHOD.DESCRIBE,
            'code': self.code,
            'model_state': model_state,
            'attribute': attribute
        }
        df = self._run_command(params)
        return df


connection_args = OrderedDict(
    code={
        'type': ARG_TYPE.PATH,
        'description': 'The path to model code'
    },
    modules={
        'type': ARG_TYPE.PATH,
        'description': 'The path to model requirements'
    }
)


from typing import List

import pandas as pd
from mindsdb_sql.parser import ast

from mindsdb.integrations.handlers.utilities.query_utilities import (
    SELECTQueryExecutor,
    SELECTQueryParser,
)
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.utilities.sql_utils import conditions_to_filter


class CustomAPITable(APITable):
    name: str = None
    columns: List[str] = [
        "category",
        "date",
        "downloads",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        """columns

        Args:
            ignore (List[str], optional): exclusion items. Defaults to [].

        Returns:
            List[str]: available columns with `ignore` items removed from the list.
        """

        return [item for item in self.columns if item not in ignore]


class PyPIRecentTable(CustomAPITable):
    name: str = "recent"
    columns: List[str] = [
        "last_day",
        "last_week",
        "last_month",
    ]

    def select(self, query: ast.Select) -> pd.DataFrame:
        """triggered at the SELECT query

        Args:
            query (ast.Select): user's entered query

        Returns:
            pd.DataFrame: the queried information
        """
        params = conditions_to_filter(query.where)

        package_name = params["package"]
        period = params.get("period", None)
        all_cols = {
            "day": "last_day",
            "week": "last_week",
            "month": "last_month",
        }

        to_be_excluded = []

        if period:
            if period in all_cols.keys():
                del all_cols[period]
                to_be_excluded = list(all_cols.values())
            else:
                raise ValueError(
                    "Make sure that one of `day`, `week` or `month` values is assigned to `period`."
                )

        select_statement_parser = SELECTQueryParser(
            query, PyPIRecentTable.name, self.get_columns(to_be_excluded)
        )
        (
            selected_columns,
            _,
            order_by_conditions,
            _,
        ) = select_statement_parser.parse_query()

        raw_df = self.handler.connection(name=package_name).recent(period)

        select_statement_executor = SELECTQueryExecutor(
            raw_df, selected_columns, [], order_by_conditions
        )

        result_df = select_statement_executor.execute_query()

        return result_df


class PyPIOverallTable(CustomAPITable):
    name: str = "overall"

    def select(self, query: ast.Select) -> pd.DataFrame:
        """triggered at the SELECT query

        Args:
            query (ast.Select): user's entered query

        Returns:
            pd.DataFrame: the queried information
        """
        params = conditions_to_filter(query.where)

        package_name = params["package"]
        mirrors = params.get("mirrors", None)

        select_statement_parser = SELECTQueryParser(
            query,
            PyPIOverallTable.name,
            self.get_columns(),
        )
        (
            selected_columns,
            _,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        raw_df = self.handler.connection(name=package_name, limit=result_limit).overall(
            mirrors=mirrors
        )

        select_statement_executor = SELECTQueryExecutor(
            raw_df, selected_columns, [], order_by_conditions
        )

        result_df = select_statement_executor.execute_query()

        return result_df


class PyPIPythonMajorTable(CustomAPITable):
    name: str = "python_major"

    def select(self, query: ast.Select) -> pd.DataFrame:
        """triggered at the SELECT query

        Args:
            query (ast.Select): user's entered query

        Returns:
            pd.DataFrame: the queried information
        """
        params = conditions_to_filter(query.where)

        package_name = params["package"]
        version = params.get("version", None)

        select_statement_parser = SELECTQueryParser(
            query,
            PyPIOverallTable.name,
            self.get_columns(),
        )
        (
            selected_columns,
            _,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        raw_df = self.handler.connection(
            name=package_name, limit=result_limit
        ).python_major(version=version)

        select_statement_executor = SELECTQueryExecutor(
            raw_df, selected_columns, [], order_by_conditions
        )

        result_df = select_statement_executor.execute_query()

        return result_df


class PyPIPythonMinorTable(CustomAPITable):
    name: str = "python_minor"

    def select(self, query: ast.Select) -> pd.DataFrame:
        """triggered at the SELECT query

        Args:
            query (ast.Select): user's entered query

        Returns:
            pd.DataFrame: the queried information
        """
        params = conditions_to_filter(query.where)

        package_name = params["package"]
        version = params.get("version", None)

        select_statement_parser = SELECTQueryParser(
            query,
            PyPIOverallTable.name,
            self.get_columns(),
        )
        (
            selected_columns,
            _,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        raw_df = self.handler.connection(
            name=package_name, limit=result_limit
        ).python_minor(version=version)

        select_statement_executor = SELECTQueryExecutor(
            raw_df, selected_columns, [], order_by_conditions
        )

        result_df = select_statement_executor.execute_query()

        return result_df


class PyPISystemTable(CustomAPITable):
    name: str = "system"

    def select(self, query: ast.Select) -> pd.DataFrame:
        """triggered at the SELECT query

        Args:
            query (ast.Select): user's entered query

        Returns:
            pd.DataFrame: the queried information
        """
        params = conditions_to_filter(query.where)

        package_name = params["package"]
        os = params.get("os", None)

        select_statement_parser = SELECTQueryParser(
            query,
            PyPIOverallTable.name,
            self.get_columns(),
        )
        (
            selected_columns,
            _,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        raw_df = self.handler.connection(name=package_name, limit=result_limit).system(
            os=os
        )

        select_statement_executor = SELECTQueryExecutor(
            raw_df, selected_columns, [], order_by_conditions
        )

        result_df = select_statement_executor.execute_query()

        return result_df



from typing import Optional

from sqlalchemy import null, func

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.context import context as ctx
import mindsdb.utilities.profiler as profiler


class PredictorRecordNotFound(Exception):
    def __init__(self, **kwargs):
        name = kwargs.get('name') or '-'
        predictor_id = kwargs.get('id') or '-'
        super().__init__(
            f"Predictor not found: name='{name}' id='{predictor_id}'"
        )


class MultiplePredictorRecordsFound(Exception):
    def __init__(self, **kwargs):
        name = kwargs.get('name') or '-'
        predictor_id = kwargs.get('id') or '-'
        super().__init__(
            f"Found multiple predictor with: name='{name}' id='{predictor_id}'"
        )


@profiler.profile()
def get_integration_record(name: str) -> db.Integration:
    company_id = ctx.company_id
    if company_id is None:
        company_id = null()

    record = (
        db.session.query(db.Integration)
        .filter_by(company_id=company_id, name=name)
        .first()
    )
    return record


@profiler.profile()
def get_project_record(name: str) -> db.Project:
    company_id = ctx.company_id
    if company_id is None:
        company_id = null()

    project_record = (
        db.session.query(db.Project)
        .filter(
            (func.lower(db.Project.name) == name)
            & (db.Project.company_id == company_id)
            & (db.Project.deleted_at == null())
        ).first()
    )
    return project_record


@profiler.profile()
def get_predictor_integration(record: db.Predictor) -> db.Integration:
    integration_record = (
        db.session.query(db.Integration)
        .filter_by(id=record.integration_id).first()
    )
    return integration_record


@profiler.profile()
def get_predictor_project(record: db.Predictor) -> db.Project:
    project_record = (
        db.session.query(db.Project)
        .filter_by(id=record.project_id).first()
    )
    return project_record


@profiler.profile()
def get_model_records(integration_id=None, active=True, deleted_at=null(),
                      project_name: Optional[str] = None, ml_handler_name: Optional[str] = None, **kwargs):
    kwargs['company_id'] = ctx.company_id
    if kwargs['company_id'] is None:
        kwargs['company_id'] = null()

    if deleted_at is not None:
        kwargs['deleted_at'] = deleted_at
    if active is not None:
        kwargs['active'] = active

    if project_name is not None:
        project_record = get_project_record(name=project_name)
        if project_record is None:
            return []
        kwargs['project_id'] = project_record.id

    if ml_handler_name is not None:
        ml_handler_record = get_integration_record(
            name=ml_handler_name
        )
        if ml_handler_record is None:
            # raise Exception(f'unknown ml handler: {ml_handler_name}')
            return []
        kwargs['integration_id'] = ml_handler_record.id

    if integration_id is not None:
        kwargs['integration_id'] = integration_id

    return (
        db.session.query(db.Predictor)
        .filter_by(**kwargs)
        .all()
    )


@profiler.profile()
def get_model_record(except_absent=False, ml_handler_name: Optional[str] = None,
                     project_name: Optional[str] = None, active: bool = True,
                     deleted_at=null(), version: Optional[int] = None, **kwargs):
    kwargs['company_id'] = ctx.company_id
    if kwargs['company_id'] is None:
        kwargs['company_id'] = null()

    kwargs['deleted_at'] = deleted_at
    if active is not None:
        kwargs['active'] = active
    if version is not None:
        kwargs['version'] = version

    if project_name is not None:
        project_record = get_project_record(name=project_name)
        if project_record is None:
            return []
        kwargs['project_id'] = project_record.id

    if ml_handler_name is not None:
        ml_handler_record = get_integration_record(
            name=ml_handler_name
        )
        if ml_handler_record is None:
            # raise Exception(f'unknown ml handler: {ml_handler_name}')
            return []
        kwargs['integration_id'] = ml_handler_record.id

    records = (
        db.session.query(db.Predictor)
        .filter_by(**kwargs)
        .all()
    )
    if len(records) > 1:
        raise MultiplePredictorRecordsFound(**kwargs)
    if len(records) == 0:
        if except_absent is True:
            raise PredictorRecordNotFound(**kwargs)
        else:
            return None
    return records[0]



import sys
import time
import threading
from typing import Optional, Callable
from concurrent.futures import ProcessPoolExecutor, Future

from pandas import DataFrame

from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.ml_task_queue.const import ML_TASK_TYPE
from mindsdb.integrations.libs.learn_process import learn_process, predict_process


def init_ml_handler(module_path):
    import importlib  # noqa

    from mindsdb.integrations.libs.learn_process import learn_process, predict_process  # noqa

    importlib.import_module(module_path)


def dummy_task():
    return None


def empty_callback(_task):
    return None


class WarmProcess:
    """ Class-wrapper for a process that persist for a long time. The process
        may be initialized with any handler requirements. Current implimentation
        is based on ProcessPoolExecutor just because of multiprocessing.pool
        produce daemon processes, which can not be used for learning. That
        bahaviour may be changed only using inheritance.
    """
    def __init__(self, initializer: Optional[Callable] = None, initargs: tuple = ()):
        """ create and init new process

            Args:
                initializer (Callable): the same as ProcessPoolExecutor initializer
                initargs (tuple): the same as ProcessPoolExecutor initargs
        """
        self.pool = ProcessPoolExecutor(1, initializer=initializer, initargs=initargs)
        self.last_usage_at = time.time()
        self._markers = set()
        # region bacause of ProcessPoolExecutor does not start new process
        # untill it get a task, we need manually run dummy task to force init.
        self.task = self.pool.submit(dummy_task)
        self._init_done = False
        self.task.add_done_callback(self._init_done_callback)
        # endregion

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        # workaround for https://bugs.python.org/issue39098
        if sys.version_info[0] == 3 and sys.version_info[1] <= 8:
            t = threading.Thread(target=self._shutdown)
            t.run()
        else:
            self.pool.shutdown(wait=False)

    def _shutdown(self):
        self.pool.shutdown(wait=True)

    def _init_done_callback(self, _task):
        """ callback for initial task
        """
        self._init_done = True

    def _update_last_usage_at_callback(self, _task):
        self.last_usage_at = time.time()

    def ready(self) -> bool:
        """ check is process ready to get a task or not

            Returns:
                bool
        """
        if self._init_done is False:
            self.task.result()
            self._init_done = True
        if self.task is None or self.task.done():
            return True
        return False

    def add_marker(self, marker: tuple):
        """ remember that that process processed task for that model

            Args:
                marker (tuple): identifier of model
        """
        if marker is not None:
            self._markers.add(marker)

    def has_marker(self, marker: tuple) -> bool:
        """ check if that process processed task for model

            Args:
                marker (tuple): identifier of model

            Returns:
                bool
        """
        if marker is None:
            return False
        return marker in self._markers

    def is_marked(self) -> bool:
        """ check if process has any marker

            Returns:
                bool
        """
        return len(self._markers) > 0

    def apply_async(self, func: Callable, *args: tuple, **kwargs: dict) -> Future:
        """ Run new task

            Args:
                func (Callable): function to run
                args (tuple): args to be passed to function
                kwargs (dict): kwargs to be passed to function

            Returns:
                Future
        """
        if not self.ready():
            raise Exception('Process task is not ready')
        self.task = self.pool.submit(
            func, *args, **kwargs
        )
        self.task.add_done_callback(self._update_last_usage_at_callback)
        self.last_usage_at = time.time()
        return self.task


def warm_function(func, context: str, *args, **kwargs):
    ctx.load(context)
    try:
        return func(*args, **kwargs)
    except Exception as e:
        raise RuntimeError(str(e)) from e


class ProcessCache:
    """ simple cache for WarmProcess-es
    """
    def __init__(self, ttl: int = 120):
        """ Args:
            ttl (int) time to live for unused process
        """
        self.cache = {}
        self._init = False
        self._lock = threading.Lock()
        self._ttl = ttl
        self._keep_alive = {}
        self._stop_event = threading.Event()
        self.cleaner_thread = None
        self._start_clean()

    def __del__(self):
        self._stop_clean()

    def _start_clean(self) -> None:
        """ start worker that close connections after ttl expired
        """
        if (
            isinstance(self.cleaner_thread, threading.Thread)
            and self.cleaner_thread.is_alive()
        ):
            return
        self._stop_event.clear()
        self.cleaner_thread = threading.Thread(target=self._clean)
        self.cleaner_thread.daemon = True
        self.cleaner_thread.start()

    def _stop_clean(self) -> None:
        """ stop clean worker
        """
        self._stop_event.set()

    def init(self):
        """ run processes for specified handlers
        """
        from mindsdb.interfaces.database.integrations import integration_controller
        preload_handlers = {}
        config = Config()
        is_cloud = config.get('cloud', False)

        if config['ml_task_queue']['type'] != 'redis':
            lightwood_handler = integration_controller.handler_modules['lightwood']
            if lightwood_handler.Handler is not None:
                preload_handlers[lightwood_handler.Handler] = 4 if is_cloud else 1

            huggingface_handler = integration_controller.handler_modules['huggingface']
            if huggingface_handler.Handler is not None:
                preload_handlers[huggingface_handler.Handler] = 1 if is_cloud else 0

            openai_handler = integration_controller.handler_modules['openai']
            if openai_handler.Handler is not None:
                preload_handlers[openai_handler.Handler] = 1 if is_cloud else 0

        with self._lock:
            if self._init is False:
                self._init = True
                for handler in preload_handlers:
                    self._keep_alive[handler.__name__] = preload_handlers[handler]
                    self.cache[handler.__name__] = {
                        'last_usage_at': time.time(),
                        'handler_module': handler.__module__,
                        'processes': [
                            WarmProcess(init_ml_handler, (handler.__module__,))
                            for _x in range(preload_handlers[handler])
                        ]
                    }

    def apply_async(self, task_type: ML_TASK_TYPE, model_id: int, payload: dict, dataframe: DataFrame = None) -> Future:
        """ run new task. If possible - do it in existing process, if not - start new one.

            Args: TODO rewrite!
                handler (object): handler class
                func (Callable): function to run
                model_marker (tuple): if any of processes processed task with same marker - new task will be sent to it
                args (tuple): args to be passed to function
                kwargs (dict): kwargs to be passed to function

            Returns:
                Future
        """
        if task_type in (ML_TASK_TYPE.LEARN, ML_TASK_TYPE.FINETUNE):
            func = learn_process
        elif task_type == ML_TASK_TYPE.PREDICT:
            func = predict_process
        else:
            raise Exception(f'Unknown ML task type: {task_type}')

        handler_module_path = payload['handler_meta']['module_path']
        handler_name = payload['handler_meta']['engine']
        model_marker = (model_id, payload['context']['company_id'])
        with self._lock:
            if handler_name not in self.cache:
                warm_process = WarmProcess(init_ml_handler, (handler_module_path,))
                self.cache[handler_name] = {
                    'last_usage_at': None,
                    'handler_module': handler_module_path,
                    'processes': [warm_process]
                }
            else:
                warm_process = None
                if model_marker is not None:
                    try:
                        warm_process = next(
                            p for p in self.cache[handler_name]['processes']
                            if p.ready() and p.has_marker(model_marker)
                        )
                    except StopIteration:
                        pass
                if warm_process is None:
                    try:
                        warm_process = next(
                            p for p in self.cache[handler_name]['processes']
                            if p.ready()
                        )
                    except StopIteration:
                        pass
                if warm_process is None:
                    warm_process = WarmProcess(init_ml_handler, (handler_module_path,))
                    self.cache[handler_name]['processes'].append(warm_process)

            task = warm_process.apply_async(warm_function, func, payload['context'], payload, dataframe)
            self.cache[handler_name]['last_usage_at'] = time.time()
            warm_process.add_marker(model_marker)
        return task

    def _clean(self) -> None:
        """ worker that stop unused processes
        """
        while self._stop_event.wait(timeout=10) is False:
            with self._lock:
                for handler_name in self.cache.keys():
                    processes = self.cache[handler_name]['processes']
                    processes.sort(key=lambda x: x.is_marked())

                    expected_count = 0
                    if handler_name in self._keep_alive:
                        expected_count = self._keep_alive[handler_name]

                    # stop processes which was used, it needs to free memory
                    for i, process in enumerate(processes):
                        if (
                            process.ready()
                            and process.is_marked()
                            and (time.time() - process.last_usage_at) > self._ttl
                        ):
                            processes.pop(i)
                            # del process
                            process.shutdown()
                            break

                    while expected_count > len(processes):
                        processes.append(
                            WarmProcess(init_ml_handler, (self.cache[handler_name]['handler_module'],))
                        )


process_cache = ProcessCache()



import json
import requests


class DockerHubClient:
    def __init__(self):
        self.auth_token = None
        self.docker_hub_base_endpoint = "https://hub.docker.com/v2/"

    def make_request(self, url, method='GET', data=None):
        if method not in ['GET', 'POST']:
            raise ValueError('Invalid HTTP request method')
        headers = {'Content-type': 'application/json'}
        if self.auth_token:
            headers['Authorization'] = 'JWT ' + self.auth_token
        request_method = getattr(requests, method.lower())
        if data and len(data) > 0:
            data = json.dumps(data)
            resp = request_method(url, data, headers=headers)
        else:
            resp = request_method(url, headers=headers)
        content = {}
        if resp.status_code == 200:
            content = {'content': json.loads(resp.content.decode()), 'code': 200}
        else:
            content = {'content': {}, 'code': resp.status_code, 'error': resp.text}
        return content

    def login(self, username=None, password=None):
        data = {'username': username, 'password': password}
        self.auth_token = None
        resp = self.make_request(self.docker_hub_base_endpoint + 'users/login/', 'POST', data)
        if resp['code'] == 200:
            self.auth_token = resp['content']['token']
        return resp

    def get_images_summary(self, namespace, repo):
        url = f'{self.docker_hub_base_endpoint}namespaces/{namespace}/repositories/{repo}/images-summary'
        return self.make_request(url)

    def get_repo_images(self, namespace, repo):
        url = f'{self.docker_hub_base_endpoint}namespaces/{namespace}/repositories/{repo}/images'
        return self.make_request(url)
      
    def get_repo_tag(self, namespace, repo, tag):
        url = f'{self.docker_hub_base_endpoint}namespaces/{namespace}/repositories/{repo}/tags/{tag}'
        return self.make_request(url)
      
    def get_repo_tags(self, namespace, repo):
        url = f'{self.docker_hub_base_endpoint}namespaces/{namespace}/repositories/{repo}/tags'
        return self.make_request(url)
    
    def get_org_settings(self, namespace):
        url = f'{self.docker_hub_base_endpoint}orgs/{namespace}/settings'
        return self.make_request(url)



import datetime as dt

from mindsdb.interfaces.storage import db

from .types import BotException


class ModelExecutor:

    def __init__(self, chat_task, model_name):
        self.chat_task = chat_task

        model = chat_task.session.model_controller.get_model(
            model_name,
            project_name=chat_task.project_name
        )
        model_record = db.Predictor.query.get(model['id'])
        integration_record = db.Integration.query.get(model_record.integration_id)

        self.model_info = {
            'model_name': model_name,
            'mode': model_record.learn_args['using'].get('mode'),
            'user_column': model_record.learn_args['using'].get('user_column'),
            'bot_column': model_record.learn_args['using'].get('assistant_column'),
            'output': model_record.to_predict[0],
            'engine': integration_record.engine,
        }

        # redefined prompt
        self.prompt = None

    def call(self, history, functions):
        model_info = self.model_info

        if model_info['mode'] != 'conversational':
            raise BotException('Not supported')

        messages = self._chat_history_to_conversation(history, model_info)

        if model_info['engine'] == 'langchain':

            tools = []
            for function in functions:
                tools.append({
                    'name': function.name,
                    'func': function.callback,
                    'description': function.description
                })

            context_list = [
                f"- Today's date is {dt.datetime.now().strftime('%Y-%m-%d')}."
                f" It must be used to understand the input date from string like 'tomorrow', 'today', 'yesterday'"
            ]
            context = '\n'.join(context_list)

            # call model
            params = {'tools': tools, 'context': context, 'prompt': self.prompt}

            predictions = self.chat_task.project_datanode.predict(
                model_name=model_info['model_name'],
                data=messages,
                params=params
            )

        elif model_info['engine'] == 'llama_index':
            predictions = self.chat_task.project_datanode.predict(
                model_name=model_info['model_name'],
                data=messages,
                params={'prompt': self.prompt}
            )

        else:
            raise BotException('Not supported')

        output_col = model_info['output']
        model_output = predictions.iloc[-1][output_col]
        return model_output

    def _chat_history_to_conversation(self, history, model_info):

        bot_username = self.chat_task.bot_params['bot_username']

        question_col = model_info['user_column']
        answer_col = model_info['bot_column']

        messages = []

        for message in history:
            text = message.text

            if text is None or text.strip() == '':
                # skip empty rows
                continue

            if message.user != bot_username:
                # create new message row
                messages.append({question_col: text, answer_col: None})
            else:
                if len(messages) == 0:
                    # add empty row
                    messages.append({question_col: None, answer_col: None})

                # update answer in previous column
                messages[-1][answer_col] = text
        return messages



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __description__ as description
from .__about__ import __version__ as version

try:
    from .autogluon_handler import AutoGluonHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "AutoGluon"
name = "autogluon"
type = HANDLER_TYPE.ML
permanent = True

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error"]



import requests
import pandas as pd
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse, HandlerResponse as Response, RESPONSE_TYPE
from .hn_table import StoriesTable, CommentsTable , HNStoriesTable ,JobStoriesTable, ShowStoriesTable

class HackerNewsHandler(APIHandler):
    """
    A class for handling connections and interactions with the Hacker News API.
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        self.base_url = 'https://hacker-news.firebaseio.com/v0'

        stories = StoriesTable(self)
        self._register_table('stories', stories)

        hnstories = HNStoriesTable(self)
        self._register_table('hnstories',hnstories)

        jobstories = JobStoriesTable(self)
        self._register_table('jobstories',jobstories)

        showstories = ShowStoriesTable(self)
        self._register_table('showstories', showstories)

        comments = CommentsTable(self)
        self._register_table('comments', comments)

    def connect(self):
        return

    def check_connection(self) -> StatusResponse:
        try:
            response = requests.get(f'{self.base_url}/maxitem.json')
            response.raise_for_status()
            return StatusResponse(True)
        except Exception as e:
            log.error(f'Error checking connection: {e}')
            return StatusResponse(False, str(e))

    def native_query(self, query_string: str = None):
        method_name, params = self.parse_native_query(query_string)

        df = self.call_hackernews_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def get_df_from_class(self, table: StoriesTable = None, limit: int = None):
        url = f'{self.base_url}/{table.json_endpoint}'
        response = requests.get(url)
        data = response.json()
        stories_data = []
        if limit is None:
            limit = len(data)
        for story_id in data[:limit]:
            url = f'{self.base_url}/item/{story_id}.json'
            response = requests.get(url)
            story_data = response.json()
            stories_data.append(story_data)
        return pd.DataFrame(stories_data, columns=table.columns)

    def call_hackernews_api(self, method_name: str = None, params: dict = None):
        story_method_handlers = {
            'get_top_stories': StoriesTable,
            'ask_hn_stories': HNStoriesTable,
            'get_job_stories': JobStoriesTable,
            'show_hn_stories': ShowStoriesTable,
        }
        if method_name in story_method_handlers:
            table = story_method_handlers[method_name]
            df = self.get_df_from_class(table)
        elif method_name == 'get_comments':
            item_id = params.get('item_id')
            url = f'{self.base_url}/item/{item_id}.json'
            response = requests.get(url)
            item_data = response.json()
            if 'kids' in item_data:
                comments_data = []
                for comment_id in item_data['kids']:
                    url = f'{self.base_url}/item/{comment_id}.json'
                    response = requests.get(url)
                    comment_data = response.json()
                    comments_data.append(comment_data)
                df = pd.DataFrame(comments_data)
            else:
                df = pd.DataFrame()
        else:
            raise ValueError(f'Unknown method_name: {method_name}')

        return df




from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import Select
from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

from .utils.helpers import *
import requests
import pandas as pd
from mindsdb_sql import parse_sql, get_lexer_parser


class EventStoreDB(DatabaseHandler):
    """
    Handler for EventStoreDB
    The handler uses the Atom Pub Over HTTP of EventStoreDB.
    This means that you need to enable AtomPuvOverHTTP on your EventStoreDB instance if you are using v20+

    Why not gRPC? At the moment we cannot use https://pypi.org/project/esdbclient/ which uses the gRPC endpoint
    because mysql-connector-python 8.0.32 requires protobuf<=3.20.3,>=3.11.0, but esdbclient
    requires protobuf 4.22.1.

    Why not TCP? At the moment https://github.com/epoplavskis/photon-pump only works in insecure mode and
    configuration is limited.

    Third reason, there is no official Python client at the moment of writing of this handler.
    But once there is better ESDB Python support for gRPC, we should move this integration from AtomPub to gRPC.
    """

    name = 'eventstoredb'
    # defaults to an insecure localhost single node
    scheme = 'http'
    host = 'localhost'
    port = '2113'
    is_connected = None
    basic_url = ""
    read_batch_size = 500  # should be adjusted based on use case
    headers = {
        'Accept': 'application/json',
        'ES-ResolveLinkTo': "True"
    }
    tlsverify = False

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        connection_data = kwargs['connection_data']
        username = connection_data.get('username')
        password = connection_data.get('password')
        self.host = connection_data.get('host')
        if connection_data.get('tls') is not None and isinstance(connection_data.get('tls'), bool) \
                and connection_data.get('tls'):
            self.scheme = 'https'
        if connection_data.get('port') is not None and isinstance(connection_data.get('port'), int):
            self.port = connection_data.get('port')
        if connection_data.get('page_size') is not None:
            if isinstance(connection_data.get('page_size'), int) and connection_data.get('page_size') > 0:
                self.read_batch_size = connection_data.get('page_size')
        if username is not None and password is not None:
            if isinstance(username, str) and isinstance(password, str):
                self.headers['authorization'] = get_auth_string(username, password)
        if connection_data.get('tlsverify') is not None and isinstance(connection_data.get('tlsverify'), bool):
            self.tlsverify = connection_data.get('tlsverify')

        self.basic_url = build_basic_url(self.scheme, self.host, self.port)

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self, **kwargs):
        if self.check_connection() == StatusResponse(True):
            self.is_connected = True
            log.logger.info(f'{self.name} connection successful!')
            return StatusResponse(True)
        log.logger.info(f'{self.name} connection could not be made.')
        return StatusResponse(False)

    def disconnect(self, **kwargs):
        return

    def check_connection(self) -> StatusResponse:
        try:
            response = requests.get(build_health_url(self.basic_url), verify=self.tlsverify)
            if response.status_code == 204:
                return StatusResponse(True)
        except Exception as e:
            log.logger.error(f'{self.name} check connection failed with: {e}!')
        return StatusResponse(False)

    def query(self, query: ASTNode) -> Response:
        if type(query) == Select:
            stream_name = query.from_table.parts[-1]  # i.e. table name
            params = {
                'embed': 'tryharder'
            }
            stream_endpoint = build_stream_url(self.basic_url, stream_name)
            response = requests.get(stream_endpoint, params=params, headers=self.headers, verify=self.tlsverify)
            entries = []
            if response is not None and response.status_code == 200:
                json_response = response.json()
                for entry in json_response["entries"]:
                    entry = entry_to_df(entry)
                    entries.append(entry)
                while True:
                    end_of_stream = True
                    if 'links' in json_response:
                        for link in json_response['links']:
                            if 'relation' in link:
                                if link['relation'] == 'next':
                                    end_of_stream = False
                                    response = requests.get(build_next_url(link['uri'], self.read_batch_size),
                                                            params=params, headers=self.headers, verify=self.tlsverify)
                                    json_response = response.json()
                                    for entry in json_response["entries"]:
                                        entry = entry_to_df(entry)
                                        entries.append(entry)
                    if end_of_stream:
                        break

            df = pd.concat(entries)

            return Response(
                RESPONSE_TYPE.TABLE,
                df
            )
        else:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message="Only 'select' queries are supported for EventStoreDB"
            )

    def native_query(self, query: str) -> Response:
        ast = self.parser(query, dialect='mindsdb')
        return self.query(ast)

    def get_tables(self) -> Response:
        """
        List all streams i.e tables
        """
        params = {
            'embed': 'tryharder'
        }
        stream_endpoint = build_streams_url(self.basic_url)
        response = requests.get(stream_endpoint, params=params, headers=self.headers, verify=self.tlsverify)
        streams = []
        if response is not None and response.status_code == 200:
            json_response = response.json()
            for entry in json_response["entries"]:
                if "title" in entry:
                    streams.append(entry["title"].split('@')[1])
            while True:
                end_of_stream = True
                if 'links' in json_response:
                    for link in json_response['links']:
                        if 'relation' in link:
                            if link['relation'] == 'next':
                                end_of_stream = False
                                response = requests.get(build_next_url(link['uri'], self.read_batch_size),
                                                        params=params, headers=self.headers, verify=self.tlsverify)
                                json_response = response.json()
                                for entry in json_response["entries"]:
                                    if "title" in entry:
                                        streams.append(entry["title"].split('@')[1])
                if end_of_stream:
                    break

        df = pd.DataFrame(streams,
                          columns=['table_name'])
        return Response(
            RESPONSE_TYPE.TABLE,
            df
        )

    def get_columns(self, table_name) -> Response:
        params = {
            'embed': 'tryharder'
        }
        stream_endpoint = build_stream_url_last_event(self.basic_url, table_name)
        response = requests.get(stream_endpoint, params=params, headers=self.headers, verify=self.tlsverify)
        entry = None
        if response is not None and response.status_code == 200:
            json_response = response.json()
            if json_response is not None and len(json_response) > 0:
                entry = entry_to_df(json_response["entries"][0])
        if entry is None:
            return Response(
                RESPONSE_TYPE.ERROR,
                "Could not retrieve JSON event data to infer column types."
            )
        data = []
        for k, v in entry.items():
            data.append([k, v.dtypes.name])
        df = pd.DataFrame(data, columns=['Field', 'Type'])
        return Response(
            RESPONSE_TYPE.TABLE,
            df
        )


def parse_sql(sql, dialect='sqlite'):
    # remove ending semicolon and spaces
    sql = re.sub(r'[\s;]+$', '', sql)

    lexer, parser = get_lexer_parser(dialect)
    tokens = lexer.tokenize(sql)
    ast = parser.parse(tokens)
    return ast



from typing import Optional
from collections import OrderedDict

import boto3

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from mindsdb.integrations.handlers.mysql_handler.mysql_handler import MySQLHandler
from mindsdb.integrations.handlers.postgres_handler.postgres_handler import PostgresHandler


class AuroraHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Amazon Aurora statements.
    """

    name = 'aurora'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        self.dialect = 'aurora'
        self.connection_data = connection_data
        self.kwargs = kwargs

        database_engine = ""
        if 'db_engine' not in self.connection_data:
            database_engine = self.get_database_engine()

        if self.connection_data['db_engine'] == 'mysql' or database_engine == 'aurora':
            self.db = MySQLHandler(
                name=name + 'mysql',
                connection_data=self.connection_data
            )
        elif self.connection_data['db_engine'] == 'postgresql' or database_engine == 'aurora-postgresql':
            self.db = PostgresHandler(
                name=name + 'postgresql',
                connection_data={key: self.connection_data[key] for key in self.connection_data if key != 'db_engine'}
            )
        else:
            raise Exception("The database engine should be either MySQL or PostgreSQL!")

    def get_database_engine(self):
        try:
            session = boto3.session.Session(
                aws_access_key_id=self.connection_data['aws_access_key_id'],
                aws_secret_access_key=self.connection_data['aws_secret_access_key']
            )

            rds = session.client('rds')

            response = rds.describe_db_clusters()

            return next(item for item in response if item["DBClusterIdentifier"] == self.connection_data['host'].split('.')[0])['Engine']
        except Exception as e:
            log.logger.error(f'Error connecting to Aurora, {e}!')
            log.logger.error('If the database engine is not provided as a parameter, please ensure that the credentials for the AWS account are passed in instead!')

    def __del__(self):
        self.db.__del__()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        return self.db.connect()

    def disconnect(self):
        """
        Close any existing connections.
        """

        return self.db.disconnect()

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        return self.db.check_connection()

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        return self.db.native_query(query)

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        return self.db.query(query)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        return self.db.get_tables()

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        return self.db.get_columns(table_name)


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Amazon Aurora DB cluster.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Amazon Aurora DB cluster.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Amazon Aurora DB cluster.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Amazon Aurora DB cluster.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the Amazon Aurora DB cluster. Must be an integer.'
    },
    db_engine={
        'type': ARG_TYPE.STR,
        'description': "The database engine of the Amazon Aurora DB cluster. This can take one of two values: 'mysql' or 'postgresql'. This parameter is optional, but if it is not provided, 'aws_access_key_id' and 'aws_secret_access_key' parameters must be provided"
    },
    aws_access_key_id={
        'type': ARG_TYPE.STR,
        'description': "The access key for the AWS account. This parameter is optional and is only required to be provided if the 'db_engine' parameter is not provided."
    },
    aws_secret_access_key={
        'type': ARG_TYPE.STR,
        'description': "The secret key for the AWS account. This parameter is optional and is only required to be provided if the 'db_engine' parameter is not provided."
    },
)

connection_args_example = OrderedDict(
    db_engine='mysql',
    host='mysqlcluster.cluster-123456789012.us-east-1.rds.amazonaws.com',
    port=3306,
    user='root',
    password='password',
    database='database'
)


import traceback


def check_auth(username, password, scramble_func, salt, company_id, config, logger=None):
    if logger is None:
        from mindsdb.api.mysql.mysql_proxy.utilities import logger

    try:
        hardcoded_user = config['auth'].get('username')
        hardcoded_password = config['auth'].get('password')
        if hardcoded_password is None:
            hardcoded_password = ''
        hardcoded_password_hash = scramble_func(hardcoded_password, salt)
        hardcoded_password = hardcoded_password.encode()

        if password is None:
            password = ''
        if isinstance(password, str):
            password = password.encode()

        if username != hardcoded_user:
            logger.warning(f'Check auth, user={username}: user mismatch')
            return {
                'success': False
            }

        if password != hardcoded_password and password != hardcoded_password_hash:
            logger.warning(f'check auth, user={username}: password mismatch')
            return {
                'success': False
            }

        logger.info(f'Check auth, user={username}: Ok')
        return {
            'success': True,
            'username': username
        }
    except Exception as e:
        logger.error(f'Check auth, user={username}: ERROR')
        logger.error(e)
        logger.error(traceback.format_exc())



import random
import time
import datetime as dt

import threading
import queue

from mindsdb.utilities.config import Config
from mindsdb.utilities.log import initialize_log
from mindsdb.utilities import log
from mindsdb.interfaces.storage import db

from mindsdb.interfaces.jobs.jobs_controller import JobsExecutor


logger = log.get_log('jobs')


def execute_async(q_in, q_out):

    while True:
        task = q_in.get()

        if task['type'] != 'task':
            return

        record_id = task['record_id']
        history_id = task['history_id']

        executor = JobsExecutor()
        try:
            executor.execute_task_local(record_id, history_id)
        except (KeyboardInterrupt, SystemExit):
            q_out.put(True)
            raise

        except Exception:
            db.session.rollback()

        q_out.put(True)


class Scheduler:
    def __init__(self, config=None):
        self.config = config

        self.q_in = queue.Queue()
        self.q_out = queue.Queue()
        self.work_thread = threading.Thread(target=execute_async, args=(self.q_in, self.q_out))
        self.work_thread.start()

    def __del__(self):
        self.stop_thread()

    def stop_thread(self):
        self.q_in.put({
            'type': 'exit'
        })

    def scheduler_monitor(self):
        check_interval = self.config.get('jobs', {}).get('check_interval', 30)

        while True:

            logger.debug('Scheduler check timetable')
            try:
                self.check_timetable()
            except (SystemExit, KeyboardInterrupt):
                raise
            except Exception as e:
                logger.error(e)

            # different instances should start in not the same time

            time.sleep(check_interval + random.randint(1, 10))

    def check_timetable(self):
        executor = JobsExecutor()

        exec_method = self.config.get('jobs', {}).get('executor', 'local')

        for record in executor.get_next_tasks():
            logger.info(f'Job execute: {record.name}({record.id})')
            self.execute_task(record.id, exec_method)

        db.session.remove()

    def execute_task(self, record_id, exec_method):

        executor = JobsExecutor()
        if exec_method == 'local':
            history_id = executor.lock_record(record_id)
            if history_id is None:
                # db.session.remove()
                logger.info(f'Unable create history record for {record_id}, is locked?')
                return

            # run in thread

            self.q_in.put({
                'type': 'task',
                'record_id': record_id,
                'history_id': history_id,
            })

            while True:
                try:
                    self.q_out.get(timeout=3)
                    break
                except queue.Empty:
                    # update last date:
                    history_record = db.JobsHistory.query.get(history_id)
                    history_record.updated_at = dt.datetime.now()
                    db.session.commit()

        else:
            # TODO add microservice mode
            raise NotImplementedError()

    def start(self):

        config = Config()
        db.init()
        initialize_log(config, 'jobs', wrap_print=True)
        self.config = config

        logger.info('Scheduler starts')

        try:
            self.scheduler_monitor()
        except (KeyboardInterrupt, SystemExit):

            self.stop_thread()
            pass


def start(verbose=False):
    scheduler = Scheduler()

    scheduler.start()


if __name__ == '__main__':
    start()



import pandas as pd
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities.log import get_log
from mindsdb_sql.parser import ast

logger = get_log("integrations.github_handler")


class GithubIssuesTable(APITable):
    """The GitHub Issue Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository issues" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            GitHub issues matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 20

        issues_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "issues":
                    next

                if an_order.field.parts[1] in ["created", "updated", "comments"]:
                    if issues_kwargs != {}:
                        raise ValueError(
                            "Duplicate order conditions found for created/updated/comments"
                        )

                    issues_kwargs["sort"] = an_order.field.parts[1]
                    issues_kwargs["direction"] = an_order.direction
                elif an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "state":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for state")
                if a_where[2] not in ["open", "closed", "all"]:
                    raise ValueError(
                        f"Unsupported where argument for state {a_where[2]}"
                    )

                issues_kwargs["state"] = a_where[2]

                continue
            if a_where[1] == "labels":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for labels")

                issues_kwargs["labels"] = a_where[2].split(",")

                continue
            if a_where[1] in ["assignee", "creator"]:
                if a_where[0] != "=":
                    raise ValueError(f"Unsupported where operation for {a_where[1]}")

                issues_kwargs[a_where[1]] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        self.handler.connect()

        github_issues_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:
                for an_issue in self.handler.connection.get_repo(
                    self.handler.repository
                ).get_issues(**issues_kwargs)[start : start + 10]:
                    if an_issue.pull_request:
                        continue

                    logger.debug(f"Processing issue {an_issue.number}")

                    github_issues_df = pd.concat(
                        [
                            github_issues_df,
                            pd.DataFrame(
                                [
                                    {
                                        "number": an_issue.number,
                                        "title": an_issue.title,
                                        "state": an_issue.state,
                                        "creator": an_issue.user.login,
                                        "closed_by": an_issue.closed_by.login
                                        if an_issue.closed_by
                                        else None,
                                        "labels": ",".join(
                                            [label.name for label in an_issue.labels]
                                        ),
                                        "assignees": ",".join(
                                            [
                                                assignee.login
                                                for assignee in an_issue.assignees
                                            ]
                                        ),
                                        "comments": an_issue.comments,
                                        "body": an_issue.body,
                                        "created": an_issue.created_at,
                                        "updated": an_issue.updated_at,
                                        "closed": an_issue.closed_at,
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_issues_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if github_issues_df.shape[0] >= total_results:
                break
            else:
                start += 10

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(github_issues_df) == 0:
            github_issues_df = pd.DataFrame([], columns=selected_columns)
        else:
            github_issues_df.columns = self.get_columns()
            for col in set(github_issues_df.columns).difference(set(selected_columns)):
                github_issues_df = github_issues_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                github_issues_df = github_issues_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        return github_issues_df

    def insert(self, query: ast.Insert):
        """Inserts data into the GitHub "Create an issue" API

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        if self.handler.connection_data.get("api_key", None) is None:
            raise ValueError(
                "Need an authenticated connection in order to insert a GitHub issue"
            )

        self.handler.connect()
        current_repo = self.handler.connection.get_repo(self.handler.repository)

        columns = [col.name for col in query.columns]

        supported_columns = {"title", "body", "assignees", "milestone", "labels"}

        if not set(columns).issubset(supported_columns):
            unsupported_columns = set(columns).difference(supported_columns)
            raise ValueError(
                "Unsupported columns for GitHub issue insert: "
                + ", ".join(unsupported_columns)
            )

        for a_row in query.values:
            insert_kwargs = {}
            a_value = dict(zip(columns, a_row))

            if a_value.get("title", None) is None:
                raise ValueError("Title parameter is required to insert a GitHub issue")

            if a_value.get("body", None):
                insert_kwargs["body"] = a_value["body"]

            if a_value.get("assignees", None):
                insert_kwargs["assignees"] = []
                for an_assignee in a_value["assignees"].split(","):
                    an_assignee = an_assignee.replace(" ", "")
                    try:
                        github_user = self.handler.connection.get_user(an_assignee)
                    except Exception as e:
                        raise ValueError(
                            f'Encountered an exception looking up assignee "{an_assignee}" in GitHub: '
                            f"{type(e).__name__} - {e}"
                        )

                    insert_kwargs["assignees"].append(github_user)

            if a_value.get("milestone", None):
                current_milestones = current_repo.get_milestones()

                found_existing_milestone = False
                for a_milestone in current_milestones:
                    if a_milestone.title == a_value["milestone"]:
                        insert_kwargs["milestone"] = a_milestone
                        found_existing_milestone = True
                        break

                if not found_existing_milestone:
                    logger.debug(
                        f"Milestone \"{a_value['milestone']}\" not found, creating it"
                    )
                    insert_kwargs["milestone"] = current_repo.create_milestone(
                        a_value["milestone"]
                    )
                else:
                    logger.debug(f"Milestone \"{a_value['milestone']}\" already exists")

            if a_value.get("labels", None):
                insert_kwargs["labels"] = []

                inserted_labels = []
                for a_label in a_value["labels"].split(","):
                    a_label = a_label.replace(" ", "")
                    inserted_labels.append(a_label)

                existing_labels = current_repo.get_labels()

                existing_labels_set = set([label.name for label in existing_labels])

                if not set(inserted_labels).issubset(existing_labels_set):
                    new_inserted_labels = set(inserted_labels).difference(
                        existing_labels_set
                    )
                    logger.debug(
                        "Inserting new labels: " + ", ".join(new_inserted_labels)
                    )
                    for a_new_label in new_inserted_labels:
                        current_repo.create_label(a_new_label, "000000")

                for a_label in existing_labels:
                    if a_label.name in inserted_labels:
                        insert_kwargs["labels"].append(a_label)

            try:
                current_repo.create_issue(a_value["title"], **insert_kwargs)
            except Exception as e:
                raise ValueError(
                    f"Encountered an exception creating an issue in GitHub: "
                    f"{type(e).__name__} - {e}"
                )

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "number",
            "title",
            "state",
            "creator",
            "closed_by",
            "labels",
            "assignees",
            "comments",
            "body",
            "created",
            "updated",
            "closed",
        ]


class GithubPullRequestsTable(APITable):
    """The GitHub Issue Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository pull requests" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            GitHub pull requests matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 20

        issues_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "pull_requests":
                    next

                if an_order.field.parts[1] in ["created", "updated", "popularity"]:
                    if issues_kwargs != {}:
                        raise ValueError(
                            "Duplicate order conditions found for created/updated/popularity"
                        )

                    issues_kwargs["sort"] = an_order.field.parts[1]
                    issues_kwargs["direction"] = an_order.direction
                elif an_order.field.parts[1] == "long_running":
                    if issues_kwargs != {}:
                        raise ValueError(
                            "Duplicate order conditions found for long_running"
                        )

                    issues_kwargs["sort"] = "long-running"
                    issues_kwargs["direction"] = an_order.direction
                elif an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "state":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for state")
                if a_where[2] not in ["open", "closed", "all"]:
                    raise ValueError(
                        f"Unsupported where argument for state {a_where[2]}"
                    )

                issues_kwargs["state"] = a_where[2]

                continue
            if a_where[1] in ["head", "base"]:
                if a_where[0] != "=":
                    raise ValueError(f"Unsupported where operation for {a_where[1]}")

                issues_kwargs[a_where[1]] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        self.handler.connect()

        github_pull_requests_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:
                for a_pull in self.handler.connection.get_repo(
                    self.handler.repository
                ).get_pulls(**issues_kwargs)[start : start + 10]:

                    github_pull_requests_df = pd.concat(
                        [
                            github_pull_requests_df,
                            pd.DataFrame(
                                [
                                    {
                                        "number": a_pull.number,
                                        "title": a_pull.title,
                                        "state": a_pull.state,
                                        "creator": a_pull.user.login,
                                        "labels": ",".join(
                                            [label.name for label in a_pull.labels]
                                        ),
                                        "milestone": a_pull.milestone.title
                                        if a_pull.milestone
                                        else None,
                                        "assignees": ",".join(
                                            [
                                                assignee.login
                                                for assignee in a_pull.assignees
                                            ]
                                        ),
                                        "reviewers": ",".join(
                                            [
                                                reviewer.login
                                                for reviewer in a_pull.requested_reviewers
                                            ]
                                        ),
                                        "teams": ",".join(
                                            [
                                                team.name
                                                for team in a_pull.requested_teams
                                            ]
                                        ),
                                        "comments": a_pull.comments,
                                        "review_comments": a_pull.review_comments,
                                        "draft": a_pull.draft,
                                        "is_merged": a_pull.merged,
                                        "mergeable": a_pull.mergeable,
                                        "mergeable_state": a_pull.mergeable_state,
                                        "merged_by": a_pull.merged_by.login
                                        if a_pull.merged_by
                                        else None,
                                        "rebaseable": a_pull.rebaseable,
                                        "body": a_pull.body,
                                        "base": a_pull.base.ref
                                        if a_pull.base
                                        else None,
                                        "head": a_pull.head.ref
                                        if a_pull.head
                                        else None,
                                        "commits": a_pull.commits,
                                        "additions": a_pull.additions,
                                        "deletions": a_pull.deletions,
                                        "changed_files": a_pull.changed_files,
                                        "created": a_pull.created_at,
                                        "updated": a_pull.updated_at,
                                        "merged": a_pull.merged_at,
                                        "closed": a_pull.closed_at,
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_pull_requests_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if github_pull_requests_df.shape[0] >= total_results:
                break
            else:
                start += 10

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(github_pull_requests_df) == 0:
            github_pull_requests_df = pd.DataFrame([], columns=selected_columns)
        else:
            github_pull_requests_df.columns = self.get_columns()
            for col in set(github_pull_requests_df.columns).difference(
                set(selected_columns)
            ):
                github_pull_requests_df = github_pull_requests_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                github_pull_requests_df = github_pull_requests_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        return github_pull_requests_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "number",
            "title",
            "state",
            "creator",
            "labels",
            "milestone",
            "assignees",
            "reviewers",
            "teams",
            "comments",
            "review_comments",
            "draft",
            "is_merged",
            "mergeable",
            "mergeable_state",
            "merged_by",
            "rebaseable",
            "body",
            "base",
            "head",
            "commits",
            "additions",
            "deletions",
            "changed_files",
            "created",
            "updated",
            "merged",
            "closed",
        ]

class GithubCommitsTable(APITable):
    """The GitHub Commits Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List commits" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            GitHub commits matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 20

        commits_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "commits":
                    next

                if an_order.field.parts[1] in ["author", "date", "message"]:
                    if commits_kwargs != {}:
                        raise ValueError(
                            "Duplicate order conditions found for author/date/message"
                        )

                    commits_kwargs["sort"] = an_order.field.parts[1]
                    commits_kwargs["direction"] = an_order.direction
                elif an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "author":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for author")
                commits_kwargs["author"] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        self.handler.connect()

        github_commits_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:

                for a_commit in self.handler.connection.get_repo(
                    self.handler.repository
                ).get_commits(**commits_kwargs)[start : start + 10]:
                    logger.debug(f"Processing commit {a_commit.sha}")

                    github_commits_df = pd.concat(
                        [
                            github_commits_df,
                            pd.DataFrame(
                                [
                                    {
                                        "sha": a_commit.sha,
                                        "author": a_commit.commit.author.name,
                                        "date": a_commit.commit.author.date,
                                        "message": a_commit.commit.message,
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_commits_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if github_commits_df.shape[0] >= total_results:
                break
            else:
                start += 10

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(github_commits_df) == 0:
            github_commits_df = pd.DataFrame([], columns=selected_columns)
        else:
            github_commits_df.columns = self.get_columns()
            for col in set(github_commits_df.columns).difference(
                set(selected_columns)
            ):
                github_commits_df = github_commits_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                github_commits_df = github_commits_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        return github_commits_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return ["sha", "author", "date", "message"]
      
class GithubReleasesTable(APITable):
    """The GitHub Releases Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository releases" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            GitHub releases matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'releases',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 20

        self.handler.connect()

        github_releases_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:

                for a_release in self.handler.connection.get_repo(
                    self.handler.repository
                ).get_releases()[start: start + 10]:

                    logger.debug(f"Processing release {a_release.id}")

                    github_releases_df = pd.concat(
                        [
                            github_releases_df,
                            pd.DataFrame(
                                [
                                    {
                                        "id": self.check_none(a_release.id),
                                        "author": self.check_none(a_release.author.login),
                                        "body": self.check_none(a_release.body),
                                        "created_at": self.check_none(str(a_release.created_at)),
                                        "html_url": self.check_none(a_release.html_url),
                                        "published_at": self.check_none(str(a_release.published_at)),
                                        "tag_name": self.check_none(a_release.tag_name),
                                        "title": self.check_none(a_release.title),
                                        "url": self.check_none(a_release.url),
                                        "zipball_url": self.check_none(a_release.zipball_url)
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_releases_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if github_releases_df.shape[0] >= total_results:
                break
            else:
                start += 10

        select_statement_executor = SELECTQueryExecutor(
            github_releases_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        github_releases_df = select_statement_executor.execute_query()

        return github_releases_df

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "id",
            "author",
            "body",
            "created_at",
            "html_url",
            "published_at",
            "tag_name",
            "title",
            "url",
            "zipball_url"
        ]
     
class GithubBranchesTable(APITable):
    """The GitHub Branches Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository branches" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            GitHub branches matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'branches',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 20

        self.handler.connect()

        github_branches_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:

                for branch in self.handler.connection.get_repo(self.handler.repository).get_branches()[start: start + 10]:
                    logger.debug(f"Processing branch {branch.name}")
                    raw_data = branch.raw_data
                    github_branches_df = pd.concat(
                        [
                            github_branches_df,
                            pd.DataFrame(
                                [
                                    {
                                        "name": self.check_none(raw_data["name"]),
                                        "url": "https://github.com/" + self.handler.repository + "/tree/" + raw_data["name"],
                                        "commit_sha": self.check_none(raw_data["commit"]["sha"]),
                                        "commit_url": self.check_none(raw_data["commit"]["url"]),
                                        "protected": self.check_none(raw_data["protected"])
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_branches_df.shape[0] >= total_results: break
            except IndexError:
                break

            if github_branches_df.shape[0] >= total_results:
                break
            else:
                start += 10

        select_statement_executor = SELECTQueryExecutor(
            github_branches_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        github_branches_df = select_statement_executor.execute_query()

        return github_branches_df

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "name",
            "url",
            "commit_sha",
            "commit_url",
            "protected"
        ]

class GithubContributorsTable(APITable):
    """The GitHub Contributors Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository contributors" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            GitHub contributors matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'contributors',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 20

        self.handler.connect()

        github_contributors_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:

                for contributor in self.handler.connection.get_repo(self.handler.repository).get_contributors()[start: start + 10]:
                    
                    raw_data = contributor.raw_data
                    github_contributors_df = pd.concat(
                        [
                            github_contributors_df,
                            pd.DataFrame(
                                [
                                    {
                                        "avatar_url": self.check_none(raw_data["avatar_url"]),
                                        "html_url": self.check_none(raw_data["html_url"]),
                                        "followers_url": self.check_none(raw_data["followers_url"]),
                                        "subscriptions_url": self.check_none(raw_data["subscriptions_url"]),
                                        "organizations_url": self.check_none(raw_data["organizations_url"]),
                                        "repos_url": self.check_none(raw_data["repos_url"]),
                                        "events_url": self.check_none(raw_data["events_url"]),
                                        "received_events_url": self.check_none(raw_data["received_events_url"]),
                                        "site_admin": self.check_none(raw_data["site_admin"]),
                                        "name": self.check_none(raw_data["name"]),
                                        "company": self.check_none(raw_data["company"]),
                                        "blog": self.check_none(raw_data["blog"]),
                                        "location": self.check_none(raw_data["location"]),
                                        "email": self.check_none(raw_data["email"]),
                                        "hireable": self.check_none(raw_data["hireable"]),
                                        "bio": self.check_none(raw_data["bio"]),
                                        "twitter_username": self.check_none(raw_data["twitter_username"]),
                                        "public_repos": self.check_none(raw_data["public_repos"]),
                                        "public_gists": self.check_none(raw_data["public_repos"]),
                                        "followers": self.check_none(raw_data["followers"]),
                                        "following": self.check_none(raw_data["following"]),
                                        "created_at": self.check_none(raw_data["created_at"]),
                                        "updated_at": self.check_none(raw_data["updated_at"])
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_contributors_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if github_contributors_df.shape[0] >= total_results:
                break
            else:
                start += 10

        select_statement_executor = SELECTQueryExecutor(
            github_contributors_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        github_contributors_df = select_statement_executor.execute_query()

        return github_contributors_df

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "avatar_url",
            "html_url",
            "followers_url",
            "subscriptions_url",
            "organizations_url",
            "repos_url",
            "events_url",
            "received_events_url",
            "site_admin",
            "name",
            "company",
            "blog",
            "location",
            "email",
            "hireable",
            "bio",
            "twitter_username",
            "public_repos",
            "public_gists",
            "followers",
            "following",
            "created_at",
            "updated_at"
        ]
        
class GithubProjectsTable(APITable):
    """The GitHub Projects Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository projects" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            GitHub projects matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'projects',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 20

        self.handler.connect()

        github_projects_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:

                for project in self.handler.connection.get_repo(self.handler.repository).get_projects()[start: start + 10]:
                    
                    raw_data = project.raw_data
                    github_projects_df = pd.concat(
                        [
                            github_projects_df,
                            pd.DataFrame(
                                [
                                    {
                                        "owner_url": self.check_none(raw_data["owner_url"]),
                                        "url": self.check_none(raw_data["url"]), 
                                        "html_url": self.check_none(raw_data["html_url"]),
                                        "columns_url": self.check_none(raw_data["columns_url"]), 
                                        "id": self.check_none(raw_data["id"]), 
                                        "node_id": self.check_none(raw_data["node_id"]), 
                                        "name": self.check_none(raw_data["name"]), 
                                        "body": self.check_none(raw_data["body"]), 
                                        "number": self.check_none(raw_data["number"]), 
                                        "state": self.check_none(raw_data["state"]), 
                                        "created_at": self.check_none(raw_data["created_at"]), 
                                        "updated_at": self.check_none(raw_data["updated_at"]),
                                        "creator_login": self.check_none(raw_data["creator"]["login"]),
                                        "creator_id": self.check_none(raw_data["creator"]["id"]),
                                        "creator_url": self.check_none(raw_data["creator"]["url"]),
                                        "creator_html_url": self.check_none(raw_data["creator"]["html_url"]),
                                        "creator_site_admin": self.check_none(raw_data["creator"]["site_admin"])
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_projects_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if github_projects_df.shape[0] >= total_results:
                break
            else:
                start += 10

        select_statement_executor = SELECTQueryExecutor(
            github_projects_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        github_projects_df = select_statement_executor.execute_query()

        return github_projects_df

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "owner_url", 
            "url",
            "html_url", 
            "columns_url", 
            "id", 
            "node_id",
            "name", 
            "body", 
            "number", 
            "state", 
            "created_at",
            "updated_at", 
            "creator_login", 
            "creator_id", 
            "creator_url",
            "creator_html_url", 
            "creator_site_admin"
        ]
      
class GithubMilestonesTable(APITable):
    """The GitHub Milestones Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitHub "List repository milestones" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            GitHub milestones matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'milestones',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 20

        self.handler.connect()

        github_milestones_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        while True:
            try:

                for milestone in self.handler.connection.get_repo(self.handler.repository).get_milestones()[start: start + 10]:
                    
                    raw_data = milestone.raw_data
                    github_milestones_df = pd.concat(
                        [
                            github_milestones_df,
                            pd.DataFrame(
                                [
                                    {
                                        "url": self.check_none(raw_data["url"]),
                                        "html_url": self.check_none(raw_data["html_url"]),
                                        "labels_url": self.check_none(raw_data["labels_url"]),
                                        "id": self.check_none(raw_data["id"]),
                                        "node_id": self.check_none(raw_data["node_id"]),
                                        "number": self.check_none(raw_data["number"]),
                                        "title": self.check_none(raw_data["title"]),
                                        "description": self.check_none(raw_data["description"]),
                                        "creator": self.check_none(raw_data["creator"]),
                                        "open_issues": self.check_none(raw_data["open_issues"]),
                                        "closed_issues": self.check_none(raw_data["closed_issues"]),
                                        "state": self.check_none(raw_data["state"]), 
                                        "created_at": self.check_none(raw_data["created_at"]),
                                        "updated_at": self.check_none(raw_data["updated_at"]),
                                        "due_on": self.check_none(raw_data["due_on"]),
                                        "closed_at": self.check_none(raw_data["closed_at"])
                                    }
                                ]
                            ),
                        ]
                    )

                    if github_milestones_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if github_milestones_df.shape[0] >= total_results:
                break
            else:
                start += 10

        select_statement_executor = SELECTQueryExecutor(
            github_milestones_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        github_milestones_df = select_statement_executor.execute_query()

        return github_milestones_df

    def check_none(self, val):
        return "" if val is None else val

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "url", 
            "html_url", 
            "labels_url", 
            "id", 
            "node_id", 
            "number", 
            "title", 
            "description", 
            "creator", 
            "open_issues", 
            "closed_issues", 
            "state", 
            "created_at", 
            "updated_at", 
            "due_on", 
            "closed_at"
        ]
